"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from typing import Any, Optional

from pygame.rect import Rect

from .constants import ColorLike, PointLike, TileFlags
from .map import TiledMap

logger = logging.getLogger(__name__)

try:
    import pygame
    from pygame._sdl2 import Renderer, Texture
except ImportError:
    logger.error("cannot import pygame (is it installed?)")
    raise


@dataclass(order=True)
class PygameSDL2Tile:
    texture: Texture
    srcrect: Rect
    size: PointLike
    angle: float = 0.0
    center: Optional[PointLike] = None
    flipx: bool = False
    flipy: bool = False


def handle_flags(flags: Optional[TileFlags]) -> tuple[float, bool, bool]:
    """
    Return angle and flip values for the SDL2 renderer
    """
    if flags is None:
        return 0.0, False, False

    if flags.flipped_diagonally:
        if flags.flipped_vertically:
            return 270.0, False, False
        else:
            return 90.0, False, False
    else:
        return 0.0, flags.flipped_horizontally, flags.flipped_vertically


def pygame_sd2_image_loader(
    renderer: Renderer,
    filename: str,
    colorkey: Optional[ColorLike] = None,
    **kwargs: Any,
) -> Callable[
    [Optional[Rect], Optional[TileFlags], Optional[PointLike]], PygameSDL2Tile
]:
    """
    pytmx image loader for pygame
    """
    image = pygame.image.load(filename)

    if colorkey:
        if isinstance(colorkey, str):
            if not colorkey.startswith("#") and len(colorkey) in (6, 8):
                colorkey = pygame.Color(f"#{colorkey}")
            else:
                colorkey = pygame.Color(colorkey)
        elif isinstance(colorkey, tuple) and 3 <= len(colorkey) <= 4:
            colorkey = pygame.Color(colorkey)
        else:
            logger.error("Invalid colorkey")
            raise ValueError("Invalid colorkey")

    parent_rect: Rect = image.get_rect()
    texture: Texture = Texture.from_surface(renderer, image)

    def load_image(
        rect: Optional[Rect] = None,
        flags: Optional[TileFlags] = None,
        center: Optional[PointLike] = None,
    ) -> PygameSDL2Tile:
        if rect:
            assert parent_rect.contains(rect), "Tile rect must be within image bounds"
        else:
            rect = parent_rect

        angle, flipx, flipy = handle_flags(flags)
        rect = Rect(*rect)
        size = rect.size

        return PygameSDL2Tile(
            texture=texture,
            srcrect=rect,
            size=size,
            angle=angle,
            center=center,
            flipx=flipx,
            flipy=flipy,
        )

    return load_image


def load_pygame_sdl2(
    renderer: Renderer, filename: str, *args: Any, **kwargs: Any
) -> TiledMap:
    """
    Load a TMX file, images, and return a TiledMap class
    """
    kwargs["image_loader"] = partial(pygame_sd2_image_loader, renderer)
    return TiledMap(filename, *args, **kwargs)



"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <https://www.gnu.org/licenses/>.

Tiled map model and image management.

This module defines `TiledMap` and related helpers used when loading TMX maps.
Imports are organized and documented for clarity.
"""

import json
import os
import warnings

# --- stdlib imports ---------------------------------------------------------
from collections import defaultdict
from collections.abc import Iterable
from itertools import chain, product
from logging import getLogger
from operator import attrgetter
from typing import Any, Iterator, Optional, Self
from xml.etree import ElementTree

# --- internal imports -------------------------------------------------------
from .class_type import TiledClassType
from .constants import GID_TRANS_ROT, MapPoint, TileFlags
from .element import TiledElement
from .group_layer import TiledGroupLayer
from .image_layer import TiledImageLayer
from .layer import TiledLayer
from .object import TiledObject
from .object_group import TiledObjectGroup
from .tile_layer import TiledTileLayer
from .tileset import TiledTileset
from .utils import decode_gid, default_image_loader

logger = getLogger(__name__)


class TiledMap(TiledElement):
    """Contains the layers, objects, and images from a Tiled .tmx map."""

    def __init__(
        self,
        filename: Optional[str] = None,
        custom_property_filename: Optional[list[str]] = None,
        image_loader=default_image_loader,
        optional_gids: Optional[set[int]] = set(),
        load_all_tiles: Optional[bool] = True,
        invert_y: Optional[bool] = True,
        allow_duplicate_names: bool = False,
    ) -> None:
        """Load new Tiled map from a .tmx file.

        Args:
            filename (Optional[str]): Filename of tiled map to load.
            custom_property_filename (Optional[list[str]]): Custom property file to load.
            image_loader (Optional[List[str]]): Function that will load images (see below).
            optional_gids (set[int]): Load specific tile image GID, even if never used.
            load_all_tiles (bool): Load all tile images, even if never used.
            invert_y (bool): Invert the y axis.
            allow_duplicate_names (bool): Allow duplicates in objects' metadata.
        """
        # allow duplicate names to be parsed and loaded
        # honor explicit constructor argument (do not read from kwargs here)
        super().__init__(allow_duplicate_names=allow_duplicate_names)

        self.filename = filename
        self.custom_property_filename = custom_property_filename
        self.image_loader = image_loader

        # optional keyword arguments checked here
        self.optional_gids = optional_gids
        self.load_all_tiles = load_all_tiles
        self.invert_y = invert_y

        # all layers in proper order
        self.layers: list[TiledLayer] = []
        # TiledTileset objects
        self.tilesets: list[TiledTileset] = []
        # tiles that have properties
        self.tile_properties: dict[int, dict[str, str]] = {}
        self.layernames: dict[str, TiledLayer] = {}
        self.objects_by_id: dict[str, TiledObject] = {}
        self.objects_by_name: dict[str, TiledObject] = {}

        # only used tiles are actually loaded, so there will be a difference
        # between the GIDs in the Tiled map data (tmx) and the data in this
        # object and the layers.  This dictionary keeps track of that.
        self.gidmap: defaultdict[int, list[tuple[int, Optional[TileFlags]]]] = (
            defaultdict(list)
        )
        # mapping of gid and trans flags to real gids
        self.imagemap: dict[tuple[int, TileFlags], tuple[int, TileFlags]] = {}
        # mapping of tiledgid to pytmx gid
        self.tiledgidmap: dict[int, int] = {}
        self.maxgid = 1

        # should be filled in by a loader function
        self.images = list()

        # defaults from the TMX specification
        self.version = "0.0"
        self.tiledversion = ""
        self.orientation = "orthogonal"
        self.renderorder = "right-down"
        self.width = 0  # width of map in tiles
        self.height = 0  # height of map in tiles
        self.tilewidth = 0  # width of a tile in pixels
        self.tileheight = 0  # height of a tile in pixels
        self.hexsidelength = 0
        self.staggeraxis = None
        self.staggerindex = None
        self.background_color = None
        self.nextobjectid = 0

        self.custom_types = dict()

        # initialize the gid mapping
        self.imagemap[(0, 0)] = 0

        if custom_property_filename:
            try:
                with open(custom_property_filename) as f:
                    self.parse_json(json.load(f))
            except (OSError, json.JSONDecodeError) as e:
                logger.error(
                    f"Error loading custom property file: {custom_property_filename}"
                )
                raise e

        if filename:
            try:
                root_node = ElementTree.parse(self.filename).getroot()
                self.parse_xml(root_node)
            except (OSError, ElementTree.ParseError) as e:
                logger.error(f"Error loading map file: {self.filename}")
                raise e

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: "{self.filename}">'

    # iterate over layers and objects in map
    def __iter__(self) -> Iterator[Self]:
        return chain(self.layers, self.objects)

    def _set_properties(
        self, node: ElementTree.Element, customs: Optional[dict[str, Any]] = None
    ) -> None:
        super()._set_properties(node, customs)

        # TODO: make class/layer-specific type casting
        # layer height and width must be int, but TiledElement.set_properties()
        # make a float by default, so recast as int here
        self.height = int(self.height)
        self.width = int(self.width)

    def parse_json(self, data: dict) -> None:
        """Parse custom data types from a JSON object

        Args:
            data (dict): Dictionary from JSON object to parse
        """
        for custom_type in data:
            if custom_type["type"] == "class":
                new = TiledClassType(custom_type["name"], custom_type["members"])

                self.custom_types[custom_type["name"]] = new

    def parse_xml(self, node: ElementTree.Element) -> Self:
        """
        Parse a TiledMap layer from ElementTree xml node.

        Returns:
            TiledMap: The parsed TiledMap layer.
        """
        self._set_properties(node)
        self.background_color = node.get("backgroundcolor", None)

        # ***         do not change this load order!         *** #
        # ***    gid mapping errors will occur if changed    *** #
        for subnode in node.findall(".//group"):
            self.add_layer(TiledGroupLayer(self, subnode))

        for subnode in node.findall(".//layer"):
            self.add_layer(TiledTileLayer(self, subnode))

        for subnode in node.findall(".//imagelayer"):
            self.add_layer(TiledImageLayer(self, subnode))

        # this will only find objectgroup layers, not including tile colliders
        for subnode in node.findall(".//objectgroup"):
            objectgroup = TiledObjectGroup(self, subnode, self.custom_types)
            self.add_layer(objectgroup)
            for obj in objectgroup:
                self.objects_by_id[obj.id] = obj
                self.objects_by_name[obj.name] = obj

        for subnode in node.findall(".//tileset"):
            self.add_tileset(TiledTileset(self, subnode))

        # "tile objects", objects with a GID, require their attributes to be
        # set after the tileset is loaded, so this step must be performed last
        # also, this step is performed for objects to load their tiles.

        # tiled stores the origin of GID objects by the lower right corner
        # this is different for all other types, so i just adjust it here
        # so all types loaded with pytmx are uniform.

        # iterate through tile objects and handle the image
        for o in [o for o in self.objects if o.gid]:
            # Decode rotation and flipping flags from the GID
            _, flags = decode_gid(o.gid)
            rotation = self.get_rotation_from_flags(
                flags
            )  # Get rotation based on flags

            # gids might also have properties assigned to them
            # in that case, assign the gid properties to the object as well
            p = self.get_tile_properties_by_gid(o.gid)
            if p:
                for key in p:
                    o.properties.setdefault(key, p[key])

            # Adjust based on rotation
            if rotation == 90:
                o.x, o.y = o.x + o.height, o.y
            elif rotation == 180:
                o.x, o.y = o.x + o.width, o.y + o.height
            elif rotation == 270:
                o.x, o.y = o.x, o.y + o.width

            # Adjust Y-coordinate if invert_y is enabled
            if self.invert_y:
                o.y -= o.height

        self.reload_images()
        return self

    def get_rotation_from_flags(self, flags: TileFlags) -> int:
        """Determine the rotation angle from TileFlags."""
        if flags.flipped_diagonally:
            if flags.flipped_horizontally and not flags.flipped_vertically:
                return 90
            elif flags.flipped_horizontally and flags.flipped_vertically:
                return 180
            elif not flags.flipped_horizontally and flags.flipped_vertically:
                return 270
        return 0

    def reload_images(self) -> None:
        """Load or reload the map images from disk.

        This method will use the image loader passed in the constructor
        to do the loading or will use a generic default, in which case no
        images will be loaded.
        """
        self.images = [None] * self.maxgid

        # iterate through tilesets to get source images
        for ts in self.tilesets:
            # skip tilesets without a source
            if ts.source is None:
                continue

            path = os.path.join(os.path.dirname(self.filename), ts.source)
            colorkey = getattr(ts, "trans", None)
            loader = self.image_loader(path, colorkey, tileset=ts)

            p = product(
                range(
                    ts.margin,
                    ts.height + ts.margin - ts.tileheight + 1,
                    ts.tileheight + ts.spacing,
                ),
                range(
                    ts.margin,
                    ts.width + ts.margin - ts.tilewidth + 1,
                    ts.tilewidth + ts.spacing,
                ),
            )

            # iterate through the tiles
            for real_gid, (y, x) in enumerate(p, ts.firstgid):
                rect = (x, y, ts.tilewidth, ts.tileheight)
                gids = self.map_gid(real_gid)

                # gids is None if the tile is never used
                # but give another chance to load the gid anyway
                if gids is None:
                    if self.load_all_tiles or real_gid in self.optional_gids:
                        # TODO: handle flags? - might never be an issue, though
                        gids = [self.register_gid(real_gid, flags=0)]

                if gids:
                    # flags might rotate/flip the image, so let the loader
                    # handle that here
                    for gid, flags in gids:
                        self.images[gid] = loader(rect, flags)
                # else:
                #     # not used in layer data give another chance to load the tile anyway
                #     if self.load_all_tiles or real_gid in self.optional_gids:
                #         # TODO: handle flags? - might never be an issue, though
                #         self.register_gid(real_gid, flags=0)

        # load image layer images
        for layer in (i for i in self.layers if isinstance(i, TiledImageLayer)):
            source = getattr(layer, "source", None)
            if source:
                colorkey = getattr(layer, "trans", None)
                real_gid = len(self.images)
                gid = self.register_gid(real_gid)
                layer.gid = gid
                path = os.path.join(os.path.dirname(self.filename), source)
                loader = self.image_loader(path, colorkey)
                image = loader()
                self.images.append(image)

        # load images in tiles.
        # instead of making a new gid, replace the reference to the tile that
        # was loaded from the tileset
        for real_gid, props in self.tile_properties.items():
            source = props.get("source", None)
            if source:
                colorkey = props.get("trans", None)
                path = os.path.join(os.path.dirname(self.filename), source)
                loader = self.image_loader(path, colorkey)
                image = loader()
                self.images[real_gid] = image

    def get_tile_image(self, x: int, y: int, target_layer: int) -> Any:
        """Return the tile image for this location.

        Args:
            x (int): The x coordinate.
            y (int): The y coordinate.
            target_layer (int): The layer's number.

        Returns:
            Any: the image object type will depend on the loader (ie. pygame surface).

        Raises:
            TypeError: if coordinates are not integers.
            ValueError: if the coordinates are out of bounds, or GID not found.
        """
        try:
            x = int(x)
            y = int(y)
            target_layer = int(target_layer)
        except ValueError:
            msg = f"Tile coordinates and layers must be integers. Got: ({x}, {y}), layer={target_layer}"
            logger.error(msg)
            raise TypeError(msg)

        if not (x >= 0 and y >= 0 and target_layer >= 0):
            msg = f"Tile coordinates and layers must be non-negative, were ({x}, {y}), layer={target_layer}"
            logger.error(msg)
            raise ValueError(msg)

        try:
            layer = self.layers[target_layer]
        except IndexError:
            msg = f"Layer {target_layer} not found"
            logger.error(msg)
            raise ValueError(msg)

        if not isinstance(layer, TiledTileLayer):
            msg = f"Layer {target_layer} is not a tile layer"
            logger.error(msg)
            raise ValueError(msg)

        try:
            gid = layer.data[y][x]
        except IndexError:
            msg = "Coords: ({x},{y}) in layer {target_layer} is invalid"
            logger.error(msg)
            raise ValueError(msg)
        except TypeError:
            msg = "Tiles must be specified in integers."
            logger.error(msg)
            raise TypeError(msg)

        else:
            return self.get_tile_image_by_gid(gid)

    def get_tile_image_by_gid(self, gid: int, *args, **kwargs) -> Any:
        """
        Return the tile image for this location.

        Args:
            gid (int): GID of image.
            *args: Additional arguments. added for compatibility with old API.
            **kwargs: Additional keyword arguments. added for compatibility with old API.

        Returns:
            Any: The image object.

        Raises:
            TypeError: If `gid` is not a non-negative integer.
            ValueError: If there is no image for this GID.
        """
        # Deprecation notice: any extra args/kwargs are ignored and will be disallowed
        # in a future release. Emit a warning to help users migrate.
        if args or kwargs:
            warnings.warn(
                "get_tile_image_by_gid(gid, *args, **kwargs): extra arguments are deprecated and ignored; pass only 'gid'",
                DeprecationWarning,
                stacklevel=2,
            )

        if not isinstance(gid, int) or gid < 0:
            msg = f"GIDs must be a non-negative integer. Got: {gid}"
            logger.debug(msg)
            raise TypeError(msg)

        try:
            return self.images[gid]
        except IndexError:
            msg = f"GID: {gid} not found"
            logger.debug(msg)
            raise ValueError(msg)

    def get_tile_gid(self, x: int, y: int, target_layer: int) -> int:
        """Return the tile image GID for this location.

        Args:
            x (int): The x coordinate.
            y (int): The y coordinate.
            target_layer (int): The layer's number.

        Returns:
            int: The tile GID.

        Raises:
            ValueError: If coordinates are out of bounds.
        """
        try:
            x = int(x)
            y = int(y)
            target_layer = int(target_layer)
        except ValueError:
            msg = f"Tile coordinates and layers must be integers. Got: ({x}, {y}), layer={target_layer}"
            logger.error(msg)
            raise TypeError(msg)

        if not (x >= 0 and y >= 0 and target_layer >= 0):
            msg = f"Tile coordinates and layers must be non-negative, were ({x}, {y}), layer={target_layer}"
            logger.error(msg)
            raise ValueError(msg)

        try:
            layer = self.layers[target_layer]
        except IndexError:
            msg = f"Layer {target_layer} not found"
            logger.error(msg)
            raise ValueError(msg)

        try:
            return layer.data[y][x]
        except IndexError:
            msg = f"Coords: ({x},{y}) in layer {target_layer} is invalid"
            logger.error(msg)
            raise ValueError(msg)

    def get_tile_properties_by_gid(self, gid: int) -> Optional[dict]:
        """Get the tile properties of a tile GID.

        Args:
            gid (int): GID.

        Returns:
            Optional[dict]: Dictionary of properties for GID, or None.
        """
        return self.tile_properties.get(gid, None)

    def get_tile_properties(self, x: int, y: int, layer: int) -> Optional[dict]:
        """Return the tile image GID for this location.

        Args:
            x (int): The x coordinate.
            y (int): The y coordinate.
            layer (int): The layer number.

        Returns:
            Optional[dict]: Dictionary of the properties for tile in this location or None.

        Raises:
            ValueError: If coordinates are out of bounds
        """
        gid = self.get_tile_gid(x, y, layer)
        return self.get_tile_properties_by_gid(gid)

    def get_tile_locations_by_gid(
        self, gid: int, only_visible: bool = True
    ) -> Iterable[MapPoint]:
        """Search map for tile locations by the GID.

        Note: Not a fast operation.  Cache results if used often.

        Args:
            gid (int): GID to be searched for.
            only_visible (bool): If True, only return locations from visible tile layers.

        Returns:
            Iterable[MapPoint]: (int, int, int) tuples, where the layer is index of the tile layers.
        """
        for l in self.visible_tile_layers if only_visible else self.tile_layers:
            for x, y, _gid in [i for i in self.layers[l].iter_data() if i[2] == gid]:
                yield x, y, l

    def set_tile_properties(self, gid: int, properties: dict) -> None:
        """Set the tile properties of a tile GID.

        Args:
            gid (int): GID.
            properties (dict): Python dictionary of properties for GID.
        """
        self.tile_properties[gid] = properties

    def get_tile_properties_by_layer(
        self, layer: int
    ) -> Iterator[tuple[int, dict[str, Any]]]:
        """Get the tile properties of each GID in layer.

        Args:
            layer (int): The layer number.

        Returns:
            Iterator[Tuple[int, dict[str, Any]]]: Iterator of GID, properties tuples for each tile in the layer.
        """
        try:
            assert int(layer) >= 0
            layer = int(layer)
        except (TypeError, AssertionError):
            msg = f"Layer must be a positive integer.  Got {type(layer)} instead."
            logger.debug(msg)
            raise ValueError(msg)

        p = product(range(self.width), range(self.height))
        layergids = set(self.layers[layer].data[y][x] for x, y in p)

        for gid in layergids:
            try:
                yield gid, self.tile_properties[gid]
            except KeyError:
                continue

    def add_layer(
        self,
        layer: TiledLayer,
    ) -> None:
        """Add a layer to the map.

        Args:
            layer (TiledLayer): The layer.
        """
        assert isinstance(
            layer, (TiledGroupLayer, TiledTileLayer, TiledImageLayer, TiledObjectGroup)
        )

        self.layers.append(layer)
        self.layernames[layer.name] = layer

    def add_tileset(self, tileset: TiledTileset) -> None:
        """Add a tileset to the map."""
        assert isinstance(tileset, TiledTileset)
        self.tilesets.append(tileset)

    def get_layer_by_name(self, name: str) -> TiledLayer:
        """Return a layer by name.

        Args:
            name (str): The layer's name. Case-sensitive!

        Returns:
            TiledLayer: The layer.

        Raises:
            ValueError: if layer by name does not exist
        """
        try:
            return self.layernames[name]
        except KeyError:
            msg = f'Layer "{name}" not found.'
            logger.debug(msg)
            raise ValueError(msg)

    def get_object_by_id(self, obj_id: int) -> TiledObject:
        """Find an object by the object id.

        Args:
            obj_id (int): ID of the object, from Tiled.

        Returns:
            TiledObject: The found object.
        """
        return self.objects_by_id[obj_id]

    def get_object_by_name(self, name: str) -> TiledObject:
        """Find an object by name, case-sensitive.

        Args:
            name (str): The object's name.

        Returns:
            TiledObject: The found object.

        Raises:
            ValueError: if object by name does not exist
        """
        try:
            return self.objects_by_name[name]
        except KeyError:
            msg = f'Object "{name}" not found.'
            logger.debug(msg)
            raise ValueError(msg)

    def get_tileset_from_gid(self, gid: int) -> TiledTileset:
        """Return tileset that owns the gid.

        Note: this is a slow operation, so if you are expecting to do this
              often, it would be worthwhile to cache the results of this.

        Args:
            gid (int): GID of tile image.

        Returns:
            TiledTileset: The tileset that owns the GID.

        Raises:
            ValueError: if the tileset for gid is not found
        """
        try:
            tiled_gid = self.tiledgidmap[gid]
        except KeyError:
            msg = f"Tile GID {gid} not found."
            logger.debug(msg)
            raise ValueError(msg)

        for tileset in sorted(self.tilesets, key=attrgetter("firstgid"), reverse=True):
            if tiled_gid >= tileset.firstgid:
                return tileset

        msg = "Tileset not found"
        logger.debug(msg)
        raise ValueError(msg)

    def get_tile_colliders(self) -> Iterable[tuple[int, list[dict]]]:
        """Return iterator of (gid, dict) pairs of tiles with colliders.

        Returns:
            Iterable[Tuple[int, List[Dict]]]: The tile colliders.
        """
        for gid, props in self.tile_properties.items():
            colliders = props.get("colliders")
            if colliders:
                yield gid, colliders

    def get_tile_flags_by_gid(self, gid: int) -> TileFlags:
        """Return the tile flags for this GID.

        Args:
            gid (int): The tile GID.

        Returns:
            TileFlags: The tile flags.
        """
        real_gid = self.tiledgidmap[gid]
        flags_list = self.gidmap[real_gid]
        for tile_gid, flags in flags_list:
            if gid == tile_gid:
                return flags

    def pixels_to_tile_pos(self, position: tuple[int, int]) -> tuple[int, int]:
        """Convert pixel position to tile position.

        Args:
            position (tuple[int, int]): The pixel position.

        Returns:
            tuple[int, int]: The tile position.
        """
        return int(position[0] / self.tilewidth), int(position[1] / self.tileheight)

    @property
    def objectgroups(self) -> Iterable[TiledObjectGroup]:
        """Returns iterator of all object groups.

        Returns:
            Iterable[TiledObjectGroup]: Iterator of all object groups.
        """
        return (layer for layer in self.layers if isinstance(layer, TiledObjectGroup))

    @property
    def objects(self) -> Iterable[TiledObject]:
        """Returns iterator of all the objects associated with the map.

        Returns:
            Iterable[TiledObject]: Iterator of all objects associated with the map.
        """
        return chain(*self.objectgroups)

    @property
    def visible_layers(self) -> Iterable[TiledLayer]:
        """Returns iterator of Layer objects that are set "visible".

        Returns:
            Iterable[TiledLayer]: Iterator of Layer objects that are set "visible".
        """

        return (l for l in self.layers if l.visible)

    @property
    def visible_tile_layers(self) -> Iterable[TiledTileLayer]:
        """Return iterator of layer indexes that are set "visible".

        Returns:
            Iterable[TiledTileLayer]: A list of layer indexes.
        """
        return (
            i
            for (i, l) in enumerate(self.layers)
            if l.visible and isinstance(l, TiledTileLayer)
        )

    @property
    def visible_object_groups(self) -> Iterable[TiledObjectGroup]:
        """Return iterator of object group indexes that are set "visible".

        Returns:
            Iterable[TiledObjectGroup]: A list of object group indexes that are set to "visible".
        """
        return (
            i
            for (i, l) in enumerate(self.layers)
            if l.visible and isinstance(l, TiledObjectGroup)
        )

    def register_gid(self, tiled_gid: int, flags: Optional[TileFlags] = None) -> int:
        """Used to manage the mapping of GIDs between .tmx and pytmx.

        Args:
            tiled_gid (int): GID that is found in TMX data.
            flags (TileFlags): TileFlags.

        Returns:
            int: New or existing GID for pytmx use.
        """
        if flags is None:
            flags = TileFlags(False, False, False)

        if tiled_gid:
            try:
                return self.imagemap[(tiled_gid, flags)][0]
            except KeyError:
                gid = self.maxgid
                self.maxgid += 1
                self.imagemap[(tiled_gid, flags)] = (gid, flags)
                self.gidmap[tiled_gid].append((gid, flags))
                self.tiledgidmap[gid] = tiled_gid
                return gid

        else:
            return 0

    def register_gid_check_flags(self, tiled_gid: int) -> int:
        """Used to manage the mapping of GIDs between .tmx and pytmx.

        Checks the GID for rotation/flip flags

        Args:
            tiled_gid (int): GID that is found in TMX data.

        Returns:
            int: New or existing GID for pytmx use.
        """
        # NOTE: the register* methods are getting really spaghetti-like
        if tiled_gid == 0:
            return 0
        elif tiled_gid < GID_TRANS_ROT:
            return self.register_gid(tiled_gid)
        else:
            return self.register_gid(*decode_gid(tiled_gid))

    def map_gid(self, tiled_gid: int) -> Optional[list[int]]:
        """Used to lookup a GID read from a TMX file's data.

        Args:
            tiled_gid (int): GID. that is found in the .tmx file data.

        Returns:
            Optional[List[int]]: List of GIDs.
        """
        try:
            return self.gidmap[int(tiled_gid)]
        except KeyError:
            return None
        except (TypeError, ValueError):
            msg = "GIDs must be an integer"
            logger.debug(msg)
            raise TypeError(msg)

    def map_gid2(self, tiled_gid: int) -> list[tuple[int, Optional[int]]]:
        """WIP.  need to refactor the gid code"""
        tiled_gid = int(tiled_gid)

        # gidmap is a default dict, so cannot trust to raise KeyError
        if tiled_gid in self.gidmap:
            return self.gidmap[tiled_gid]
        else:
            gid = self.register_gid(tiled_gid)
            return [(gid, None)]



"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <https://www.gnu.org/licenses/>.

Tiled object model and parser.
"""

from typing import TYPE_CHECKING, Any, Self
from xml.etree import ElementTree

from .constants import Point
from .element import TiledElement
from .utils import rotate

if TYPE_CHECKING:
    from .map import TiledMap


class TiledObject(TiledElement):
    """
    Represents any Tiled Object.

    Supported types: Box, Ellipse, Tile Object, Polyline, Polygon, Text, Point.
    """

    def __init__(
        self,
        parent: "TiledMap",
        node: ElementTree.Element,
        custom_types: dict[str, Any],
    ) -> None:
        super().__init__()
        self.parent = parent

        # defaults from the specification
        self.id = 0
        self.name = None
        self.type = None
        self.object_type = "rectangle"
        self.x = 0
        self.y = 0
        self.width = 0
        self.height = 0
        self.rotation = 0
        self.gid = 0
        self.visible = 1
        self.closed = True
        self.template = None
        self.custom_types = custom_types

        self.parse_xml(node)

    @property
    def image(self) -> Any:
        """Image for the object, if assigned.

        Returns:
            Any: The image object type will depend on the loader (ie. pygame.Surface).
        """
        if self.gid:
            return self.parent.images[self.gid]
        return None

    def parse_xml(self, node: ElementTree.Element) -> Self:
        """
        Parse a TiledObject layer from ElementTree xml node.

        Returns:
            TiledObject: The parsed TiledObject layer.
        """

        def read_points(text) -> tuple[tuple[float, float]]:
            """Parse a text string of float tuples and return [(x,...),...]"""
            return tuple(tuple(map(float, i.split(","))) for i in text.split())

        self._set_properties(node, self.custom_types)

        # Handle tile objects
        if self.gid:
            self.object_type = "tile"
            self.gid = self.parent.register_gid_check_flags(self.gid)

        points = None
        node_handlers = {
            "polygon": {
                "type": "polygon",
                "points_attr": "points",
                "parse": read_points,
                "closed": True,
            },
            "polyline": {
                "type": "polyline",
                "points_attr": "points",
                "parse": read_points,
                "closed": False,
            },
            "ellipse": {"type": "ellipse"},
            "point": {"type": "point"},
            "text": {"type": "text"},
        }

        for node_name, handler in node_handlers.items():
            subnode = node.find(node_name)
            if subnode is not None:
                self.object_type = handler["type"]

                if node_name == "text":
                    # Inline text property parsing
                    setattr(self, "text", subnode.text)
                    setattr(
                        self, "font_family", subnode.get("fontfamily", "Sans Serif")
                    )
                    setattr(self, "pixel_size", int(subnode.get("pixelsize", 16)))
                    setattr(self, "wrap", bool(subnode.get("wrap", False)))
                    setattr(self, "bold", bool(subnode.get("bold", False)))
                    setattr(self, "italic", bool(subnode.get("italic", False)))
                    setattr(self, "underline", bool(subnode.get("underline", False)))
                    setattr(self, "strike_out", bool(subnode.get("strikeout", False)))
                    setattr(self, "kerning", bool(subnode.get("kerning", True)))
                    setattr(self, "h_align", subnode.get("halign", "left"))
                    setattr(self, "v_align", subnode.get("valign", "top"))
                    setattr(self, "color", subnode.get("color", "#000000FF"))

                if "points_attr" in handler:
                    points = handler["parse"](subnode.get(handler["points_attr"]))
                if "closed" in handler:
                    self.closed = handler["closed"]
                break

        if points:
            xs, ys = zip(*points)
            self.width = max(xs) - min(xs)
            self.height = max(ys) - min(ys)
            self.points = tuple([Point(i[0] + self.x, i[1] + self.y) for i in points])
        elif self.object_type == "rectangle":
            self.points = tuple(
                [
                    Point(self.x, self.y),
                    Point(self.x + self.width, self.y),
                    Point(self.x + self.width, self.y + self.height),
                    Point(self.x, self.y + self.height),
                ]
            )

        return self

    def apply_transformations(self) -> list[Point]:
        """Return all points for object, taking in account rotation."""
        if hasattr(self, "points"):
            return rotate(self.points, self, self.rotation)
        else:
            return rotate(self.as_points, self, self.rotation)

    @property
    def as_points(self) -> list[Point]:
        return [
            Point(*i)
            for i in [
                (self.x, self.y),
                (self.x, self.y + self.height),
                (self.x + self.width, self.y + self.height),
                (self.x + self.width, self.y),
            ]
        ]



import logging
import unittest

from pytmx.element import TiledElement
from pytmx.map import TiledMap
from pytmx.utils import convert_to_bool

# Tiled gid flags
GID_TRANS_FLIPX = 1 << 31
GID_TRANS_FLIPY = 1 << 30
GID_TRANS_ROT = 1 << 29
GID_MASK = GID_TRANS_FLIPX | GID_TRANS_FLIPY | GID_TRANS_ROT


class TestConvertToBool(unittest.TestCase):
    def test_string_string_true(self) -> None:
        self.assertTrue(convert_to_bool("1"))
        self.assertTrue(convert_to_bool("y"))
        self.assertTrue(convert_to_bool("Y"))
        self.assertTrue(convert_to_bool("t"))
        self.assertTrue(convert_to_bool("T"))
        self.assertTrue(convert_to_bool("yes"))
        self.assertTrue(convert_to_bool("Yes"))
        self.assertTrue(convert_to_bool("YES"))
        self.assertTrue(convert_to_bool("true"))
        self.assertTrue(convert_to_bool("True"))
        self.assertTrue(convert_to_bool("TRUE"))

    def test_string_string_false(self) -> None:
        self.assertFalse(convert_to_bool("0"))
        self.assertFalse(convert_to_bool("n"))
        self.assertFalse(convert_to_bool("N"))
        self.assertFalse(convert_to_bool("f"))
        self.assertFalse(convert_to_bool("F"))
        self.assertFalse(convert_to_bool("no"))
        self.assertFalse(convert_to_bool("No"))
        self.assertFalse(convert_to_bool("NO"))
        self.assertFalse(convert_to_bool("false"))
        self.assertFalse(convert_to_bool("False"))
        self.assertFalse(convert_to_bool("FALSE"))

    def test_string_number_true(self) -> None:
        self.assertTrue(convert_to_bool(1))
        self.assertTrue(convert_to_bool(1.0))

    def test_string_number_false(self) -> None:
        self.assertFalse(convert_to_bool(0))
        self.assertFalse(convert_to_bool(0.0))
        self.assertFalse(convert_to_bool(-1))
        self.assertFalse(convert_to_bool(-1.1))

    def test_string_bool_true(self) -> None:
        self.assertTrue(convert_to_bool(True))

    def test_string_bool_false(self) -> None:
        self.assertFalse(convert_to_bool(False))

    def test_string_bool_none(self) -> None:
        self.assertFalse(convert_to_bool(None))

    def test_string_bool_empty(self) -> None:
        self.assertFalse(convert_to_bool(""))

    def test_string_bool_whitespace_only(self) -> None:
        self.assertFalse(convert_to_bool(" "))

    def test_non_boolean_string_raises_error(self) -> None:
        with self.assertRaises(ValueError):
            convert_to_bool("garbage")

    def test_non_boolean_number_raises_error(self) -> None:
        with self.assertRaises(ValueError):
            convert_to_bool("200")

    def test_edge_cases(self):
        # Whitespace
        self.assertTrue(convert_to_bool("  t  "))
        self.assertFalse(convert_to_bool("  f  "))

        # Numeric edge cases
        self.assertTrue(convert_to_bool(1e-10))  # Very small positive number
        self.assertFalse(convert_to_bool(-1e-10))  # Very small negative number


class TiledMapTest(unittest.TestCase):
    filename = "tests/resources/test01.tmx"

    def setUp(self) -> None:
        self.m = TiledMap(self.filename)

    def test_build_rects(self) -> None:
        try:
            from pytmx import util_pygame

            rects = util_pygame.build_rects(self.m, "Grass and Water", "tileset", None)
            self.assertEqual(rects[0], [0, 0, 240, 240])
            rects = util_pygame.build_rects(self.m, "Grass and Water", "tileset", 18)
            self.assertNotEqual(0, len(rects))
        except ImportError:
            pass

    def test_get_tile_image(self) -> None:
        image = self.m.get_tile_image(0, 0, 0)

    def test_get_tile_image_by_gid(self) -> None:
        image = self.m.get_tile_image_by_gid(0)
        self.assertIsNone(image)

        image = self.m.get_tile_image_by_gid(1)
        self.assertIsNotNone(image)

    def test_reserved_names_check_disabled_with_option(self) -> None:
        tiled_map = TiledMap(allow_duplicate_names=True)
        items = [("name", "conflict")]
        self.assertFalse(tiled_map._contains_invalid_property_name(items))

    def test_map_width_height_is_int(self) -> None:
        self.assertIsInstance(self.m.width, int)
        self.assertIsInstance(self.m.height, int)

    def test_layer_width_height_is_int(self) -> None:
        self.assertIsInstance(self.m.layers[0].width, int)
        self.assertIsInstance(self.m.layers[0].height, int)

    def test_properties_are_converted_to_builtin_types(self) -> None:
        self.assertIsInstance(self.m.properties["test_bool"], bool)
        self.assertIsInstance(self.m.properties["test_color"], str)
        self.assertIsInstance(self.m.properties["test_file"], str)
        self.assertIsInstance(self.m.properties["test_float"], float)
        self.assertIsInstance(self.m.properties["test_int"], int)
        self.assertIsInstance(self.m.properties["test_string"], str)

    def test_properties_are_converted_to_correct_values(self) -> None:
        self.assertFalse(self.m.properties["test_bool"])
        self.assertTrue(self.m.properties["test_bool_true"])

    def test_pixels_to_tile_pos(self) -> None:
        self.assertEqual(self.m.pixels_to_tile_pos((0, 33)), (0, 2))
        self.assertEqual(self.m.pixels_to_tile_pos((33, 0)), (2, 0))
        self.assertEqual(self.m.pixels_to_tile_pos((0, 0)), (0, 0))
        self.assertEqual(self.m.pixels_to_tile_pos((65, 86)), (4, 5))



"""
This is tested on pygame 2.0.1 and python 3.9.6.
Leif Theden "bitcraft", 2012-2025

Rendering demo for the TMXLoader.

"""

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import pygame
from pygame._sdl2 import Renderer, Window
from pygame.locals import *

from pytmx import __version__
from pytmx.layer import TiledTileLayer
from pytmx.util_pygame_sdl2 import load_pygame_sdl2

logger = logging.getLogger(__name__)


@dataclass
class GameContext:
    window: Window
    renderer: Renderer


class TiledRenderer:
    """
    Super simple way to render a tiled map
    """

    def __init__(self, ctx: GameContext, filename: str) -> None:
        self.ctx = ctx
        self.tmx_data = tm = load_pygame_sdl2(ctx.renderer, filename)
        self.pixel_size = tm.width * tm.tilewidth, tm.height * tm.tileheight

    def render_map(self) -> None:
        """
        Render our map to a pygame surface

        Feel free to use this as a starting point for your pygame app.

        Scrolling is a often requested feature, but pytmx is a map
        loader, not a renderer!  If you'd like to have a scrolling map
        renderer, please see my pyscroll project.
        """
        # iterate over all the visible layers, then draw them
        for layer in self.tmx_data.visible_layers:
            if isinstance(layer, TiledTileLayer):
                self.render_tile_layer(layer)

    def render_tile_layer(self, layer: TiledTileLayer) -> None:
        """
        Render all TiledTiles in this layer
        """
        tw = self.tmx_data.tilewidth
        th = self.tmx_data.tileheight

        # iterate over the tiles in the layer, and draw them
        for x, y, image in layer.tiles():
            x *= tw
            y *= th
            image.texture.draw(
                image.srcrect,
                (x, y, tw, th),
                image.angle,
                image.center,
                image.flipx,
                image.flipy,
            )


class SimpleTest:
    """
    Basic app to display a rendered Tiled map
    """

    def __init__(self, ctx: GameContext, filename: Path) -> None:
        self.ctx = ctx
        self.map_renderer: Optional[TiledRenderer] = None
        self.running: bool = False
        self.exit_status: int = 0
        self.load_map(filename)

    def load_map(self, filename: Path) -> None:
        """
        Create a renderer, load data, and print some debug info
        """
        self.map_renderer = TiledRenderer(self.ctx, filename.as_posix())

        logger.info("Objects in map:")
        for obj in self.map_renderer.tmx_data.objects:
            logger.info("Object: %s", obj)
            logger.debug("Properties: %s", vars(obj))

        logger.info("GID (tile) properties:")
        for k, v in self.map_renderer.tmx_data.tile_properties.items():
            logger.info("%s\t%s", k, v)

        logger.info("Tile colliders:")
        for k, v in self.map_renderer.tmx_data.get_tile_colliders():
            logger.info("%s\t%s", k, list(v))

    def draw(self) -> None:
        """
        Draw our map to some surface (probably the display)
        """
        if self.map_renderer:
            self.map_renderer.render_map()
            self.ctx.renderer.present()

    def handle_input(self) -> None:
        try:
            event = pygame.event.wait()

            if event.type == QUIT:
                self.exit_status = 0
                self.running = False

            elif event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    self.exit_status = 0
                    self.running = False
                else:
                    self.running = False

        except KeyboardInterrupt:
            self.exit_status = 0
            self.running = False

    def run(self) -> int:
        """
        Main loop of the app.

        Returns:
            int: Exit status (0 = success, 1 = error)
        """
        self.running = True
        self.exit_status = 1

        while self.running:
            self.handle_input()
            self.draw()

        return self.exit_status


if __name__ == "__main__":
    pygame.init()
    pygame.font.init()
    window = Window("pytmx map viewer", size=(600, 600))
    renderer = Renderer(window, vsync=True)

    ctx = GameContext(window, renderer)

    logging.basicConfig(level=logging.DEBUG)
    logger.info(__version__)

    # loop through a bunch of maps in the maps folder
    try:
        for filename in Path("apps/data").glob("*.tmx"):
            logger.info("Testing %s", filename)
            renderer.clear()
            if not SimpleTest(ctx, filename).run():
                break
    except Exception as e:
        logger.exception("Unhandled exception: %s", e)
        pygame.quit()
        raise



"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <https://www.gnu.org/licenses/>.
"""

from typing import TYPE_CHECKING, Any, Self
from xml.etree import ElementTree

from .element import TiledElement

if TYPE_CHECKING:
    from .map import TiledMap


class TiledImageLayer(TiledElement):
    """Represents Tiled Image Layer.

    The image associated with this layer will be loaded and assigned a GID.
    """

    def __init__(self, parent: "TiledMap", node: ElementTree.Element) -> None:
        super().__init__()
        self.parent = parent
        self.source = None
        self.trans = None
        self.gid = 0

        # defaults from the specification
        self.name = None
        self.opacity = 1
        self.visible = 1

        self.parse_xml(node)

    @property
    def image(self) -> Any:
        """Image for the object, if assigned.

        Returns:
            Any: the image object type will depend on the loader (ie. pygame.Surface).
        """
        if self.gid:
            return self.parent.images[self.gid]
        return None

    def parse_xml(self, node: ElementTree.Element) -> Self:
        """
        Parse a TiledImageLayer layer from ElementTree xml node.

        Returns:
            TiledImageLayer: The parsed TiledImageLayer layer.
        """
        self._set_properties(node)
        self.name = node.get("name", None)
        self.opacity = node.get("opacity", self.opacity)
        self.visible = node.get("visible", self.visible)
        image_node = node.find("image")
        self.source = image_node.get("source", None)
        self.trans = image_node.get("trans", None)
        return self



"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <https://www.gnu.org/licenses/>.

Object group model and parser.
"""

from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Optional, Self, Union
from xml.etree import ElementTree

from .element import TiledElement
from .object import TiledObject

if TYPE_CHECKING:
    from .map import TiledMap


class TiledObjectGroup(TiledElement):
    """
    Represents a Tiled ObjectGroup, acting as a container (list) for TiledObjects.
    """

    def __init__(
        self,
        parent: "TiledMap",
        node: ElementTree.Element,
        custom_types: Optional[dict[str, Any]] = None,
    ) -> None:
        super().__init__()
        self.parent = parent

        self._objects: list[TiledObject] = []

        # defaults from the specification
        self.name = None
        self.color = None
        self.opacity = 1
        self.visible = 1
        self.offsetx = 0
        self.offsety = 0
        self.custom_types = custom_types
        self.draworder = "index"

        self.parse_xml(node)

    def parse_xml(self, node: ElementTree.Element) -> Self:
        """
        Parse a TiledObjectGroup layer from ElementTree xml node.

        Returns:
            TiledObjectGroup: The parsed TiledObjectGroup layer.
        """
        self._set_properties(node, self.custom_types)

        self._objects.extend(
            TiledObject(self.parent, child, self.custom_types)
            for child in node.findall("object")
        )

        return self

    def __len__(self) -> int:
        """Returns the number of objects in the group."""
        return len(self._objects)

    def __iter__(self) -> Iterator[TiledObject]:
        """Allows iteration (e.g., for obj in group:)."""
        return iter(self._objects)

    def __getitem__(
        self, index: Union[int, slice]
    ) -> Union[TiledObject, list[TiledObject]]:
        """Allows indexing and slicing (e.g., group[0], group[2:5])."""
        return self._objects[index]

    def append(self, obj: TiledObject) -> None:
        self._objects.append(obj)

    def remove(self, obj: TiledObject) -> None:
        """Remove a specific object from the group."""
        self._objects.remove(obj)

    def clear(self) -> None:
        """Remove all objects from the group."""
        self._objects.clear()

    def find_by_name(self, name: str) -> Optional[TiledObject]:
        """Find the first object with a matching name."""
        for obj in self._objects:
            if getattr(obj, "name", None) == name:
                return obj
        return None



from typing import Union

from .group_layer import TiledGroupLayer
from .image_layer import TiledImageLayer
from .object_group import TiledObjectGroup
from .tile_layer import TiledTileLayer

TiledLayer = Union[TiledTileLayer, TiledImageLayer, TiledGroupLayer, TiledObjectGroup]



"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <https://www.gnu.org/licenses/>.

Tiled Tileset parser and model.
"""

import logging
import os
from typing import TYPE_CHECKING, Any, Self
from xml.etree import ElementTree
from xml.etree.ElementTree import ParseError

from .constants import AnimationFrame
from .element import TiledElement
from .object_group import TiledObjectGroup
from .properties import parse_properties, types

if TYPE_CHECKING:
    from .map import TiledMap

logger = logging.getLogger(__name__)


class TiledTileset(TiledElement):
    """Represents a Tiled Tileset

    External tilesets are supported.  GID/ID's from Tiled are not
    guaranteed to be the same after loaded.
    """

    def __init__(self, parent: "TiledMap", node: ElementTree.Element) -> None:
        """Represents a Tiled Tileset

        Args:
            parent (TiledMap): The parent TiledMap.
            node (ElementTree.Element): The XML node.
        """
        super().__init__()
        self.parent = parent
        self.offset = (0, 0)
        self.tileset_source = None

        # defaults from the specification
        self.firstgid = 0
        self.source = None
        self.name = None
        self.tilewidth = 0
        self.tileheight = 0
        self.spacing = 0
        self.margin = 0
        self.tilecount = 0
        self.columns = 0

        # image properties
        self.trans = None
        self.width = 0
        self.height = 0

        self.parse_xml(node)

    def _resolve_path(self, path: str, relative_to_source: bool) -> str:
        """
        Resolve a path relative to either the TMX or TSX file, but keep it relative.
        """
        base = os.path.dirname(
            self.tileset_source
            if relative_to_source and self.tileset_source
            else self.parent.filename
        )
        resolved = os.path.join(base, path)
        logger.debug(f"Resolved path: {resolved}")
        return resolved

    def _parse_tile_properties(self, node: ElementTree.Element) -> dict[str, Any]:
        """
        Parses a single tile's attributes and custom properties.
        """
        props = {k: types[k](v) for k, v in node.items()}
        props.update(parse_properties(node))
        logger.debug(f"Parsed tile properties: {props}")
        return props

    def _parse_animation_frames(
        self, anim_node: ElementTree.Element
    ) -> list[AnimationFrame]:
        """
        Parses animation frames from a tile's animation node.
        """
        frames = []
        for frame in anim_node.findall("frame"):
            duration = int(frame.get("duration"))
            gid = self.parent.register_gid(int(frame.get("tileid")) + self.firstgid)
            frames.append(AnimationFrame(gid, duration))
            logger.debug(
                f"Parsed animation frame: tileid={frame.get('tileid')}, duration={duration}, gid={gid}"
            )
        return frames

    def _handle_external_source(self, node: ElementTree.Element) -> ElementTree.Element:
        """Handles parsing an external tileset source."""
        source = node.get("source", None)
        if not source:
            return node

        if source[-4:].lower() != ".tsx":
            msg = f"Found external tileset, but cannot handle type: {source}"
            logger.error(msg)
            raise ValueError(msg)

        self.tileset_source = source
        self.firstgid = int(node.get("firstgid"))
        logger.debug(f"External tileset detected: {source}, firstgid={self.firstgid}")

        resolved_path = self._resolve_path(source, relative_to_source=False)

        try:
            new_node = ElementTree.parse(resolved_path).getroot()
            logger.debug(f"Successfully loaded external tileset from {resolved_path}")
            return new_node
        except FileNotFoundError as e:
            msg = f"Cannot find tileset file {source} from {self.parent.filename}, should be at {resolved_path}"
            logger.error(msg)
            raise FileNotFoundError(msg) from e
        except ParseError as e:
            msg = f"Error loading external tileset: {resolved_path}"
            logger.error(msg)
            raise ParseError(msg) from e

    def _parse_all_tiles(self, node: ElementTree.Element, is_external: bool) -> None:
        """Parses all individual tiles within the tileset node."""
        for child in node.iter("tile"):
            tiled_gid = int(child.get("id"))
            props = self._parse_tile_properties(child)
            logger.debug(f"Parsing tile ID: {tiled_gid}")

            image_node = child.find("image")
            if image_node is not None:
                tile_source = image_node.get("source")
                if is_external:
                    tile_source = self._resolve_path(
                        tile_source, relative_to_source=True
                    )
                props["source"] = tile_source
                props["trans"] = image_node.get("trans", None)
                props["width"] = int(image_node.get("width"))
                props["height"] = int(image_node.get("height"))
                logger.debug(
                    f"Tile image parsed: source={tile_source}, size={props['width']}x{props['height']}"
                )
            else:
                props["width"] = self.tilewidth
                props["height"] = self.tileheight

            anim = child.find("animation")
            props["frames"] = (
                self._parse_animation_frames(anim) if anim is not None else []
            )

            for objgrp_node in child.findall("objectgroup"):
                objectgroup = TiledObjectGroup(self.parent, objgrp_node, None)
                props["colliders"] = objectgroup
                logger.debug(f"Object group parsed for tile ID {tiled_gid}")

            for gid, flags in self.parent.map_gid2(tiled_gid + self.firstgid):
                self.parent.set_tile_properties(gid, props)

    def _parse_tileset_image_and_offset(self, node: ElementTree.Element) -> None:
        """Parses the main tileset image and offset nodes."""
        tile_offset_node = node.find("tileoffset")
        if tile_offset_node is not None:
            self.offset = (
                int(tile_offset_node.get("x", 0)),
                int(tile_offset_node.get("y", 0)),
            )
            logger.debug(f"Parsed tileoffset: {self.offset}")

        image_node = node.find("image")
        if image_node is not None:
            self.source = image_node.get("source")
            if self.tileset_source:
                self.source = self._resolve_path(self.source, relative_to_source=True)
            self.trans = image_node.get("trans", None)
            self.width = int(image_node.get("width"))
            self.height = int(image_node.get("height"))
            logger.debug(
                f"Tileset image node parsed: source={self.source}, size={self.width}x{self.height}, trans={self.trans}"
            )

    def parse_xml(self, node: ElementTree.Element) -> Self:
        """
        Parse a TiledTileset layer from ElementTree xml node.

        Returns:
            TiledTileset: The parsed TiledTileset layer.
        """
        logger.debug("Starting XML parsing for tileset")

        original_source = node.get("source")

        # Handle external tileset source
        node = self._handle_external_source(node)

        # Set main tileset properties
        self._set_properties(node)
        logger.debug(
            f"Tileset properties set: name={self.name}, tilecount={self.tilecount}, columns={self.columns}"
        )

        # Parse all individual tiles
        self._parse_all_tiles(node, bool(original_source))

        # Parse main tileset image and offset
        self._parse_tileset_image_and_offset(node)

        logger.debug("Finished parsing tileset XML")
        return self



"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <https://www.gnu.org/licenses/>.

Constants and simple data structures for pytmx.

This module centralizes bit flags, GID masks, and lightweight namedtuples
so they can be reused across the package without circular imports.

All symbols here are intentionally dependency-free.
"""

from __future__ import annotations

from typing import NamedTuple, Union

try:
    import pygame

    HAS_PYGAME = True
except ImportError:
    HAS_PYGAME = False


# --- Bit flags --------------------------------------------------------------
# Internal transform flags used by pytmx
TRANS_FLIPX = 1
TRANS_FLIPY = 2
TRANS_ROT = 4

# Tiled GID transform flags (as stored in TMX data)
GID_TRANS_FLIPX = 1 << 31
GID_TRANS_FLIPY = 1 << 30
GID_TRANS_ROT = 1 << 29
GID_MASK = GID_TRANS_FLIPX | GID_TRANS_FLIPY | GID_TRANS_ROT


# --- Lightweight named tuples -----------------------------------------------------
class AnimationFrame(NamedTuple):
    gid: int
    duration: int


class Point(NamedTuple):
    x: float
    y: float


class TileFlags(NamedTuple):
    flipped_diagonally: bool
    flipped_horizontally: bool
    flipped_vertically: bool


flag_cache: dict[int, TileFlags] = {}

# Commonly reused values
empty_flags = TileFlags(False, False, False)

# --- Shared typing aliases (kept simple here to avoid imports) --------------
ColorLike = Union[tuple[int, int, int, int], tuple[int, int, int], int, str]
MapPoint = tuple[int, int, int]
# internal flags
# error message format strings go here


if HAS_PYGAME:
    PointLike = Union[tuple[int, int], pygame.Vector2, Point]
else:
    PointLike = Union[tuple[int, int], Point]



import unittest
from unittest.mock import MagicMock, Mock, patch

import pygame
from pygame.rect import Rect
from pygame.surface import Surface

from pytmx.constants import TileFlags
from pytmx.layer import TiledTileLayer
from pytmx.map import TiledMap
from pytmx.tileset import TiledTileset
from pytmx.util_pygame import (
    build_rects,
    handle_transformation,
    pygame_image_loader,
    simplify,
    smart_convert,
)


class TestPyTMXPygameIntegration(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pygame.init()

    @classmethod
    def tearDownClass(cls):
        pygame.quit()

    def setUp(self):
        self.screen = pygame.display.set_mode((1, 1))

    def test_handle_transformation_no_flags(self):
        tile = Surface((32, 32))
        flags = MagicMock(
            flipped_diagonally=False,
            flipped_horizontally=False,
            flipped_vertically=False,
        )
        transformed_tile = handle_transformation(tile, flags)
        self.assertEqual(transformed_tile, tile)

    def test_handle_transformation_flipped_diagonally(self):
        tile = Surface((32, 32))
        flags = MagicMock(
            flipped_diagonally=True,
            flipped_horizontally=False,
            flipped_vertically=False,
        )
        transformed_tile = handle_transformation(tile, flags)
        self.assertNotEqual(transformed_tile, tile)

    def test_handle_transformation_flipped_horizontally(self):
        tile = Surface((32, 32))
        flags = MagicMock(
            flipped_diagonally=False,
            flipped_horizontally=True,
            flipped_vertically=False,
        )
        transformed_tile = handle_transformation(tile, flags)
        self.assertNotEqual(transformed_tile, tile)

    def test_handle_transformation_flipped_vertically(self):
        tile = Surface((32, 32))
        flags = MagicMock(
            flipped_diagonally=False,
            flipped_horizontally=False,
            flipped_vertically=True,
        )
        transformed_tile = handle_transformation(tile, flags)
        self.assertNotEqual(transformed_tile, tile)

    def test_smart_convert_no_colorkey(self):
        original = Surface((32, 32))
        colorkey = None
        pixelalpha = False
        converted_tile = smart_convert(original, colorkey, pixelalpha)
        self.assertEqual(converted_tile.get_flags() & pygame.SRCALPHA, 0)

    def test_smart_convert_colorkey(self):
        original = Surface((32, 32))
        colorkey = (255, 255, 255)
        pixelalpha = False
        converted_tile = smart_convert(original, colorkey, pixelalpha)
        self.assertEqual(converted_tile.get_flags() & pygame.SRCALPHA, 0)

    def test_smart_convert_pixelalpha(self):
        original = Surface((32, 32), pygame.SRCALPHA)
        colorkey = None
        pixelalpha = True
        converted_tile = smart_convert(original, colorkey, pixelalpha)
        self.assertEqual(converted_tile.get_flags() & pygame.SRCALPHA, pygame.SRCALPHA)

    @patch("pygame.image.load")
    def test_pygame_image_loader(self, mock_load):
        filename = "test_image.png"
        mock_load.return_value = Surface((32, 32))
        load_image_func = pygame_image_loader(filename, None)
        tile = load_image_func()
        self.assertIsInstance(tile, Surface)

    @patch("pygame.image.load")
    def test_pygame_image_loader_rect(self, mock_load):
        filename = "test_image.png"
        mock_load.return_value = Surface((32, 32))
        load_image_func = pygame_image_loader(filename, None)
        rect = pygame.Rect(0, 0, 16, 16)
        tile = load_image_func(rect=rect)
        self.assertIsInstance(tile, Surface)

    @patch("pygame.image.load")
    def test_pygame_image_loader_flags(self, mock_load):
        filename = "test_image.png"
        mock_load.return_value = Surface((32, 32))
        load_image_func = pygame_image_loader(filename, None)
        flags = MagicMock(
            flipped_diagonally=True, flipped_horizontally=True, flipped_vertically=True
        )
        tile = load_image_func(flags=flags)
        self.assertIsInstance(tile, Surface)

    def test_invalid_colorkey_type(self):
        with self.assertRaises(ValueError):
            pygame_image_loader("fake.png", {"not": "valid"})

    @patch("pygame.image.load")
    def test_colorkey_hex_without_hash(self, mock_load):
        mock_load.return_value = Surface((32, 32))
        loader = pygame_image_loader("fake.png", "ff00ff")
        self.assertTrue(callable(loader))

    @patch("pygame.image.load")
    def test_pygame_image_loader_out_of_bounds(self, mock_load):
        mock_load.return_value = Surface((32, 32))
        loader = pygame_image_loader("fake.png", None)
        with self.assertRaises(ValueError):
            loader(pygame.Rect(100, 100, 10, 10))

    @patch("pygame.image.load")
    def test_pygame_image_loader_rect_and_flags(self, mock_load):
        surface = Surface((64, 64))
        surface.fill((255, 255, 255))
        mock_load.return_value = surface
        flags = MagicMock(
            flipped_diagonally=True, flipped_horizontally=True, flipped_vertically=True
        )
        loader = pygame_image_loader("fake.png", None)
        tile = loader(rect=pygame.Rect(0, 0, 32, 32), flags=flags)
        self.assertIsInstance(tile, Surface)

    def test_smart_convert_preserve_alpha_flag(self):
        surface = Surface((32, 32), pygame.SRCALPHA)
        surface.fill((255, 255, 255, 255))
        tile = smart_convert(surface, None, False, preserve_alpha_flag=True)
        self.assertTrue(tile.get_flags() & pygame.SRCALPHA)


class TestSimplify(unittest.TestCase):
    def test_empty_input(self):
        self.assertEqual(simplify([], 1, 1), [])

    def test_single_point(self):
        points = [(1, 1)]
        expected_rects = [Rect(1, 1, 1, 1)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_adjacent_points(self):
        points = [(1, 1), (2, 1), (1, 2), (2, 2)]
        expected_rects = [Rect(1, 1, 2, 2)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_non_adjacent_points(self):
        points = [(1, 1), (3, 3), (5, 5)]
        expected_rects = [Rect(1, 1, 1, 1), Rect(3, 3, 1, 1), Rect(5, 5, 1, 1)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_rectangles(self):
        points = [(1, 1), (2, 1), (3, 1), (1, 2), (2, 2), (3, 2)]
        expected_rects = [Rect(1, 1, 3, 2)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_large_input(self):
        points = [(x, y) for x in range(10) for y in range(10)]
        expected_rects = [Rect(0, 0, 10, 10)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_tile_size(self):
        points = [(1, 1), (2, 1), (1, 2), (2, 2)]
        expected_rects = [Rect(2, 2, 4, 4)]
        self.assertEqual(simplify(points, 2, 2), expected_rects)

    def test_diagonal_points(self):
        points = [(1, 1), (2, 2), (3, 3)]
        expected_rects = [Rect(1, 1, 1, 1), Rect(2, 2, 1, 1), Rect(3, 3, 1, 1)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_vertical_line(self):
        points = [(1, 1), (1, 2), (1, 3), (1, 4)]
        expected_rects = [Rect(1, 1, 1, 4)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_horizontal_line(self):
        points = [(1, 1), (2, 1), (3, 1), (4, 1)]
        expected_rects = [Rect(1, 1, 4, 1)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_l_shape(self):
        points = [(1, 1), (2, 1), (3, 1), (3, 2), (3, 3)]
        expected_rects = [Rect(1, 1, 3, 1), Rect(3, 2, 1, 2)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_t_shape(self):
        points = [(1, 1), (2, 1), (3, 1), (2, 2), (2, 3)]
        expected_rects = [Rect(1, 1, 3, 1), Rect(2, 2, 1, 2)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_cross_shape(self):
        points = [(1, 1), (2, 1), (3, 1), (2, 2), (2, 3), (1, 2), (3, 2)]
        expected_rects = [Rect(1, 1, 3, 2), Rect(2, 3, 1, 1)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_large_l_shape(self):
        points = [
            (1, 1),
            (2, 1),
            (3, 1),
            (4, 1),
            (5, 1),
            (5, 2),
            (5, 3),
            (5, 4),
            (5, 5),
        ]
        expected_rects = [Rect(1, 1, 5, 1), Rect(5, 2, 1, 4)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_large_t_shape(self):
        points = [
            (1, 1),
            (2, 1),
            (3, 1),
            (4, 1),
            (5, 1),
            (3, 2),
            (3, 3),
            (3, 4),
            (3, 5),
        ]
        expected_rects = [Rect(1, 1, 5, 1), Rect(3, 2, 1, 4)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_overlapping_segments(self):
        points = [(1, 1), (2, 1), (2, 2), (3, 2)]
        expected_rects = [Rect(1, 1, 2, 1), Rect(2, 2, 2, 1)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_hollow_rectangle(self):
        points = [
            (1, 1),
            (2, 1),
            (3, 1),
            (1, 2),
            (3, 2),
            (1, 3),
            (2, 3),
            (3, 3),
        ]
        expected_rects = [
            Rect(1, 1, 3, 1),
            Rect(1, 2, 1, 2),
            Rect(2, 3, 2, 1),
            Rect(3, 2, 1, 1),
        ]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_staggered_columns(self):
        points = [(1, 1), (2, 2), (3, 3)]
        expected_rects = [Rect(1, 1, 1, 1), Rect(2, 2, 1, 1), Rect(3, 3, 1, 1)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)

    def test_scaled_large_block(self):
        points = [(x, y) for x in range(5) for y in range(5)]
        expected_rects = [Rect(0, 0, 10, 10)]
        self.assertEqual(simplify(points, 2, 2), expected_rects)

    def test_row_with_gaps(self):
        points = [(1, 1), (2, 1), (4, 1), (5, 1)]
        expected_rects = [Rect(1, 1, 2, 1), Rect(4, 1, 2, 1)]
        self.assertEqual(simplify(points, 1, 1), expected_rects)


class TestBuildRects(unittest.TestCase):

    def setUp(self):
        self.tmxmap = Mock(spec=TiledMap)
        self.tmxmap.width = 10
        self.tmxmap.height = 10
        self.tmxmap.tilewidth = 32
        self.tmxmap.tileheight = 32
        self.tmxmap.tilesets = [
            Mock(spec=TiledTileset),
            Mock(spec=TiledTileset),
        ]
        self.tmxmap.tilesets[0].name = "tileset1"
        self.tmxmap.tilesets[1].name = "tileset2"
        self.tmxmap.layers = [
            Mock(spec=TiledTileLayer),
            Mock(spec=TiledTileLayer),
        ]
        self.tmxmap.layers[0].name = "layer1"
        self.tmxmap.layers[0].data = [[1 for _ in range(10)] for _ in range(10)]
        self.tmxmap.layers[1].name = "layer2"
        self.tmxmap.layers[1].data = [[1 for _ in range(10)] for _ in range(10)]
        self.tmxmap.get_layer_data = Mock(
            return_value=[[1 for _ in range(10)] for _ in range(10)]
        )
        self.tmxmap.map_gid = Mock(return_value=[(1, 0)])

    def test_build_rects_with_int_layer_and_int_tileset(self):
        result = build_rects(self.tmxmap, 0, 0, 1)
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], Rect)

    def test_build_rects_with_str_layer_and_str_tileset(self):
        result = build_rects(self.tmxmap, "layer1", "tileset1", 1)
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], Rect)

    def test_build_rects_with_invalid_layer(self):
        with self.assertRaises(ValueError):
            build_rects(self.tmxmap, "invalid_layer", 0, 1)

    def test_build_rects_with_invalid_tileset(self):
        with self.assertRaises(IndexError):
            build_rects(self.tmxmap, 0, 2, 1)

    def test_build_rects_with_invalid_tileset_type(self):
        with self.assertRaises(ValueError):
            build_rects(self.tmxmap, 0, "invalid_type", 1)

    def test_build_rects_with_invalid_gid(self):
        self.tmxmap.map_gid = Mock(return_value=[])
        result = build_rects(self.tmxmap, 0, 0, 1)
        self.assertIsInstance(result, list)

    def test_build_rects_with_no_gid(self):
        result = build_rects(self.tmxmap, 0, 0, None)
        self.assertIsInstance(result, list)
        self.assertIsInstance(result[0], Rect)

    def test_layer_with_no_name(self):
        self.tmxmap.layers[0].name = None
        with self.assertRaises(ValueError):
            build_rects(self.tmxmap, "layer1", 0, 1)

    def test_tileset_with_no_name(self):
        self.tmxmap.tilesets[0].name = None
        with self.assertRaises(ValueError):
            build_rects(self.tmxmap, "layer1", "tileset1", 1)

    def test_gid_not_in_layer_data(self):
        self.tmxmap.map_gid = Mock(return_value=[(999, 0)])
        result = build_rects(self.tmxmap, "layer1", 0, 999)
        self.assertEqual(result, [])

    def test_sparse_tile_distribution(self):
        data = [[0 for _ in range(10)] for _ in range(10)]
        data[1][1] = 1
        data[5][5] = 1
        self.tmxmap.layers[0].data = data
        result = build_rects(self.tmxmap, "layer1", 0, None)
        self.assertTrue(len(result) >= 2)

    def test_build_rects_without_tileset(self):
        result = build_rects(self.tmxmap, "layer1", None, None)
        self.assertIsInstance(result, list)


class TestHandleTransformation(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pygame.init()

    @classmethod
    def tearDownClass(cls):
        pygame.quit()

    def setUp(self):
        self.tile = pygame.Surface((32, 32))
        self.tile.fill((255, 0, 0))

    def test_no_flags(self):
        flags = TileFlags(False, False, False)
        new_tile = handle_transformation(self.tile, flags)
        self.assertEqual(new_tile.get_size(), self.tile.get_size())

    def test_flipped_diagonally(self):
        flags = TileFlags(False, False, True)
        new_tile = handle_transformation(self.tile, flags)
        self.assertEqual(new_tile.get_size(), self.tile.get_size())

    def test_flipped_diagonally_non_square(self):
        non_square_tile = pygame.Surface((32, 64))
        non_square_tile.fill((255, 0, 0))
        flags = TileFlags(True, False, False)
        with self.assertRaises(ValueError):
            handle_transformation(non_square_tile, flags)

    def test_flipped_diagonally_square(self):
        square_tile = pygame.Surface((64, 64))
        square_tile.fill((0, 255, 0))
        flags = TileFlags(True, False, False)
        transformed_tile = handle_transformation(square_tile, flags)
        self.assertEqual(transformed_tile.get_size(), (64, 64))

    def test_flipped_horizontally(self):
        flags = TileFlags(True, False, False)
        new_tile = handle_transformation(self.tile, flags)
        self.assertEqual(new_tile.get_size(), self.tile.get_size())

    def test_flipped_vertically(self):
        flags = TileFlags(False, True, False)
        new_tile = handle_transformation(self.tile, flags)
        self.assertEqual(new_tile.get_size(), self.tile.get_size())

    def test_flipped_both(self):
        flags = TileFlags(True, True, False)
        new_tile = handle_transformation(self.tile, flags)
        self.assertEqual(new_tile.get_size(), self.tile.get_size())

    def test_flipped_all(self):
        flags = TileFlags(True, True, True)
        new_tile = handle_transformation(self.tile, flags)
        self.assertEqual(new_tile.get_size(), self.tile.get_size())


class TestSmartConvert(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pygame.init()
        cls.screen = pygame.display.set_mode((1, 1))  # Needed for convert()
        cls.red = (255, 0, 0, 255)
        cls.transparent = (0, 0, 0, 0)

        cls.base_surface = pygame.Surface((32, 32))
        cls.base_surface.fill(cls.red)

        cls.alpha_surface = pygame.Surface((32, 32), pygame.SRCALPHA)
        cls.alpha_surface.fill((255, 0, 0, 128))  # semi-transparent

        cls.edge_transparency_surface = pygame.Surface((32, 32), pygame.SRCALPHA)
        cls.edge_transparency_surface.fill((255, 255, 255, 255))
        for x in range(32):
            cls.edge_transparency_surface.set_at((x, 0), cls.transparent)

    @classmethod
    def tearDownClass(cls):
        pygame.quit()

    def test_no_colorkey(self):
        result = smart_convert(self.base_surface, None, False)
        self.assertEqual(result.get_size(), self.base_surface.get_size())
        self.assertFalse(result.get_flags() & pygame.SRCALPHA)

    def test_pixelalpha_enabled(self):
        result = smart_convert(self.alpha_surface, None, True)
        self.assertTrue(result.get_flags() & pygame.SRCALPHA)

    def test_colorkey_basic(self):
        result = smart_convert(self.base_surface, self.red, False)
        self.assertEqual(result.get_colorkey(), self.red)

    def test_colorkey_rleaccel(self):
        surface = pygame.Surface((32, 32))
        surface.fill(self.red)
        result = smart_convert(surface, self.red, False)
        self.assertEqual(result.get_colorkey(), self.red)
        self.assertIn(result.get_flags() & pygame.RLEACCEL, [0, pygame.RLEACCEL])

    def test_sparse_colorkey(self):
        surface = pygame.Surface((32, 32))
        surface.fill((0, 0, 255))
        surface.set_at((0, 0), self.red)
        result = smart_convert(surface, self.red, False)
        self.assertEqual(result.get_colorkey(), self.red)
        self.assertFalse(result.get_flags() & pygame.RLEACCEL)

    def test_edge_transparency_handling(self):
        result = smart_convert(self.edge_transparency_surface, None, True)
        self.assertTrue(result.get_flags() & pygame.SRCALPHA)

    def test_colorkey_with_pixelalpha_enabled(self):
        result = smart_convert(self.base_surface, self.red, True)
        self.assertEqual(result.get_colorkey(), self.red)

    @patch("pygame.mask.from_surface")
    def test_mask_failure_fallback(self, mock_mask):
        # Simulate pygame raising an error during mask generation; util should fallback
        mock_mask.side_effect = pygame.error("Fake mask failure")
        result = smart_convert(self.base_surface, None, False)
        self.assertTrue(result.get_flags() & pygame.SRCALPHA)

    def test_returned_surface_format(self):
        result = smart_convert(self.base_surface, None, False)
        self.assertFalse(result.get_flags() & pygame.SRCALPHA)

    def test_colorkey_and_pixelalpha_combinations(self):
        test_cases = [
            {"colorkey": None, "pixelalpha": False},
            {"colorkey": None, "pixelalpha": True},
            {"colorkey": self.red, "pixelalpha": False},
            {"colorkey": self.red, "pixelalpha": True},
        ]

        for case in test_cases:
            with self.subTest(case=case):
                result = smart_convert(
                    self.base_surface, case["colorkey"], case["pixelalpha"]
                )
                self.assertEqual(result.get_size(), self.base_surface.get_size())

    def test_conversion_time_large_surface(self):
        import timeit

        large = pygame.Surface((512, 512))
        large.fill(self.red)

        # Wrap smart_convert call in a lambda for timing
        duration = timeit.timeit(lambda: smart_convert(large, None, False), number=10)
        print(f"Conversion time for large surface (512x512): {duration:.4f} seconds")
        self.assertLess(duration, 1.0)  # Arbitrary performance threshold

    def test_batch_tile_conversion(self):
        tiles = [pygame.Surface((32, 32)) for _ in range(50)]
        for tile in tiles:
            tile.fill(self.red)
        results = [smart_convert(tile, self.red, False) for tile in tiles]
        for i, result in enumerate(results):
            self.assertEqual(result.get_size(), tiles[i].get_size())

    def test_gradient_alpha_surface(self):
        surface = pygame.Surface((32, 32), pygame.SRCALPHA)
        for y in range(32):
            alpha = int(255 * (y / 31))  # vertical gradient
            for x in range(32):
                surface.set_at((x, y), (255, 0, 0, alpha))
        result = smart_convert(surface, None, True)
        self.assertTrue(result.get_flags() & pygame.SRCALPHA)

    def test_colorkey_near_match(self):
        surface = pygame.Surface((32, 32))
        surface.fill((255, 0, 1))  # almost red
        result = smart_convert(surface, (255, 0, 0), False)
        self.assertNotEqual(result.get_colorkey(), (255, 0, 0))  # should not match

    def test_zero_sized_surface(self):
        surface = pygame.Surface((0, 0))
        result = smart_convert(surface, None, False)
        self.assertEqual(result.get_size(), (0, 0))

    def test_preserve_alpha_flag_enabled(self):
        surface = pygame.Surface((32, 32), pygame.SRCALPHA)
        surface.fill((255, 255, 255, 255))  # fully opaque
        result = smart_convert(surface, None, False, preserve_alpha_flag=True)
        self.assertTrue(result.get_flags() & pygame.SRCALPHA)

    def test_conversion_consistency(self):
        surface = pygame.Surface((32, 32))
        surface.fill((123, 123, 123))
        result1 = smart_convert(surface, None, False)
        result2 = smart_convert(surface, None, False)
        self.assertEqual(result1.get_flags(), result2.get_flags())



# -*- coding: utf-8 -*-
"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <http://www.gnu.org/licenses/>.
"""
import itertools
import logging
from collections.abc import Callable
from typing import Any, Optional, Union

from .constants import ColorLike, PointLike, TileFlags
from .map import TiledMap

logger = logging.getLogger(__name__)

try:
    import pygame
    from pygame.transform import flip, rotate
except ImportError:
    logger.error("cannot import pygame (is it installed?)")
    raise

__all__ = ["load_pygame", "pygame_image_loader", "simplify", "build_rects"]


def handle_transformation(tile: pygame.Surface, flags: TileFlags) -> pygame.Surface:
    """
    Transform tile according to the flags and return a new one

    Parameters:
        tile: tile surface to transform
        flags: TileFlags object

    Returns:
        new tile surface
    """
    if flags.flipped_diagonally:
        if tile.get_width() != tile.get_height():
            raise ValueError(
                f"Cannot flip tile {tile.get_size()} diagonally if it is not a square"
            )
        tile = flip(rotate(tile, 270), True, False)
    if flags.flipped_horizontally or flags.flipped_vertically:
        tile = flip(tile, flags.flipped_horizontally, flags.flipped_vertically)
    return tile


def count_colorkey_pixels(surface: pygame.Surface, colorkey: ColorLike) -> int:
    """Efficiently count pixels matching the colorkey."""
    try:
        import pygame.surfarray

        pixel_array = pygame.surfarray.pixels3d(surface)
        r, g, b = colorkey[:3]
        return (
            (pixel_array[:, :, 0] == r)
            & (pixel_array[:, :, 1] == g)
            & (pixel_array[:, :, 2] == b)
        ).sum()
    except ImportError:
        # Slow fallback method
        width, height = surface.get_size()
        return sum(
            1
            for x in range(width)
            for y in range(height)
            if surface.get_at((x, y))[:3] == colorkey[:3]
        )


def has_transparency(surface: pygame.Surface, threshold: int = 254) -> bool:
    """Detects transparency via mask."""
    try:
        mask = pygame.mask.from_surface(surface, threshold)
        return mask.count() < surface.get_width() * surface.get_height()
    except (pygame.error, AttributeError, TypeError, ValueError):
        return True  # Assume transparency if mask fails


def log_surface_properties(surface: pygame.Surface, label: str = "Surface") -> None:
    """Print diagnostic info (optional for debugging)."""
    size = surface.get_size()
    flags = surface.get_flags()
    bitsize = surface.get_bitsize()
    alpha = surface.get_alpha()
    logger.info(
        f"[{label}] Size: {size}, Flags: {flags}, Bitsize: {bitsize}, Alpha: {alpha}"
    )


def smart_convert(
    original: pygame.Surface,
    colorkey: Optional[ColorLike],
    pixelalpha: bool,
    preserve_alpha_flag: bool = False,
) -> pygame.Surface:
    """
    Return new pygame Surface with optimal pixel/data format

    This method does several interactive tests on a surface to determine the optimal
    flags and pixel format for each tile surface.

    Parameters:
        original: tile surface to inspect
        colorkey: optional colorkey for the tileset image
        pixelalpha: if true, prefer per-pixel alpha surfaces
        preserve_alpha_flag: if True, retain SRCALPHA format even if transparency isn't detected

    Returns:
        new tile surface
    """
    width, height = original.get_size()
    tile = None

    def force_alpha() -> pygame.Surface:
        return original.convert_alpha()

    if colorkey:
        colorkey_pixels = count_colorkey_pixels(original, colorkey)
        ratio = colorkey_pixels / (width * height)

        if ratio > 0.5:
            tile = original.convert()
            tile.set_colorkey(colorkey, pygame.RLEACCEL)
        else:
            tile = force_alpha() if pixelalpha else original.convert()
            tile.set_colorkey(colorkey)
    else:
        if has_transparency(original):
            tile = force_alpha()
        elif preserve_alpha_flag and original.get_flags() & pygame.SRCALPHA:
            tile = force_alpha()
        else:
            tile = original.convert()

    # Optional: log_surface_properties(tile, label="Converted Tile")
    return tile


def pygame_image_loader(
    filename: str, colorkey: Optional[ColorLike], **kwargs: Any
) -> Callable[[Optional[pygame.Rect], Optional[TileFlags]], pygame.Surface]:
    """
    pytmx image loader for pygame

    Parameters:
        filename: filename, including path, to load
        colorkey: colorkey for the image

    Returns:
        function to load tile images
    """
    pixelalpha = kwargs.get("pixelalpha", False)

    if colorkey:
        if isinstance(colorkey, str):
            if not colorkey.startswith("#") and len(colorkey) in (6, 8):
                colorkey = pygame.Color(f"#{colorkey}")
            else:
                colorkey = pygame.Color(colorkey)
        elif isinstance(colorkey, tuple) and 3 <= len(colorkey) <= 4:
            colorkey = pygame.Color(colorkey)
        else:
            logger.error("Invalid colorkey")
            raise ValueError("Invalid colorkey")

    image = pygame.image.load(filename)

    def load_image(
        rect: Optional[pygame.Rect] = None, flags: Optional[TileFlags] = None
    ) -> pygame.Surface:
        if rect:
            try:
                tile = image.subsurface(rect)
            except ValueError as e:
                msg = f"Tile bounds outside bounds of tileset image: {e}"
                logger.error(msg)
                raise ValueError(msg) from e
        else:
            tile = image.copy()

        if flags:
            tile = handle_transformation(tile, flags)

        tile = smart_convert(tile, colorkey, pixelalpha)
        return tile

    return load_image


def load_pygame(filename: str, *args: Any, **kwargs: Any) -> TiledMap:
    """Load a TMX file, images, and return a TiledMap class

    PYGAME USERS: Use me.

    this utility has 'smart' tile loading.  by default any tile without
    transparent pixels will be loaded for quick blitting.  if the tile has
    transparent pixels, then it will be loaded with per-pixel alpha.  this is
    a per-tile, per-image check.

    if a color key is specified as an argument, or in the tmx data, the
    per-pixel alpha will not be used at all. if the tileset's image has colorkey
    transparency set in Tiled, the util_pygame will return images that have their
    transparency already set.

    TL;DR:
    Don't attempt to convert() or convert_alpha() the individual tiles.  It is
    already done for you.

    Parameters:
        filename: filename to load

    Returns:
        new pytmx.TiledMap object
    """
    kwargs["image_loader"] = pygame_image_loader
    return TiledMap(filename, *args, **kwargs)


def build_rects(
    tmxmap: TiledMap,
    layer: Union[int, str],
    tileset: Optional[Union[int, str]],
    real_gid: Optional[int],
) -> list[pygame.Rect]:
    """
    Generate a set of non-overlapping rects that represents the distribution of the specified gid.

    Useful for generating rects for use in collision detection

    GID Note: You will need to add 1 to the GID reported by Tiled.

    Parameters:
        tmxmap: TiledMap object
        layer: int or string name of layer
        tileset: int or string name of tileset
        real_gid: Tiled GID of the tile + 1 (see note)

    Returns:
        list of pygame Rect objects
    """
    if isinstance(tileset, int):
        try:
            tileset_obj = tmxmap.tilesets[tileset]
        except IndexError:
            msg = f"Tileset #{tileset} not found in map {tmxmap}."
            logger.debug(msg)
            raise IndexError(msg)

    elif isinstance(tileset, str):
        try:
            # Find the tileset with the matching name
            tileset_obj = next((t for t in tmxmap.tilesets if t.name == tileset), None)
        except (AttributeError, TypeError) as e:
            msg = f"Error finding tileset: {e}"
            logger.debug(msg)
            raise ValueError(msg)

        if tileset_obj is None:
            msg = f'Tileset "{tileset}" not found in map {tmxmap}.'
            logger.debug(msg)
            raise ValueError(msg)

    gid = None
    if real_gid:
        try:
            # Get the map GID and flags
            map_gid = tmxmap.map_gid(real_gid)
            if map_gid:
                gid, flags = map_gid[0]
        except IndexError:
            msg = f"GID #{real_gid} not found"
            logger.debug(msg)
            raise ValueError(msg)

    if isinstance(layer, int):
        layer_data = tmxmap.get_layer_data(layer)
    elif isinstance(layer, str):
        try:
            # Find the layer with the matching name
            layer_obj = next(
                (l for l in tmxmap.layers if l.name and l.name == layer), None
            )
        except (AttributeError, TypeError) as e:
            msg = f"Error finding layer: {e}"
            logger.debug(msg)
            raise ValueError(msg)

        if layer_obj is None:
            msg = f'Layer "{layer}" not found in map {tmxmap}.'
            logger.debug(msg)
            raise ValueError(msg)

        layer_data = layer_obj.data

    points = []
    for x, y in itertools.product(range(tmxmap.width), range(tmxmap.height)):
        tile_gid = layer_data[y][x]
        if (gid is None and tile_gid) or (gid == tile_gid):
            points.append((x, y))

    rects = simplify(points, tmxmap.tilewidth, tmxmap.tileheight)
    return rects


def simplify(
    all_points: list[PointLike],
    tilewidth: int,
    tileheight: int,
) -> list[pygame.Rect]:
    """Given a list of points, return list of rects that represent them
    kludge:

    "A kludge (or kluge) is a workaround, a quick-and-dirty solution,
    a clumsy or inelegant, yet effective, solution to a problem, typically
    using parts that are cobbled together."

    -- wikipedia

    turn a list of points into a rects
    adjacent rects will be combined.

    plain english:
        the input list must be a list of tuples that represent
        the areas to be combined into rects
        the rects will be blended together over solid groups

        so if data is something like:

        0 1 1 1 0 0 0
        0 1 1 0 0 0 0
        0 0 0 0 0 4 0
        0 0 0 0 0 4 0
        0 0 0 0 0 0 0
        0 0 1 1 1 1 1

        you'll have the 4 rects that mask the area like this:

        ..######......
        ..####........
        ..........##..
        ..........##..
        ..............
        ....##########

        pretty cool, right?

    there may be cases where the number of rectangles is not as low as possible,
    but I haven't found that it is excessively bad.  certainly much better than
    making a list of rects, one for each tile on the map!
    """

    if not all_points:
        return []

    point_set = set(all_points)

    rect_list: list[pygame.Rect] = []

    def pick_rect(points: set[PointLike], rects: list[pygame.Rect]) -> None:
        """
        Recursively pick a rect from the points and add it to the rects list.
        """
        if not points:
            return

        ox, oy = min(points, key=lambda p: (p[0], p[1]))
        x = ox
        y = oy
        ex = None

        while True:
            x += 1
            if (x, y) not in points:
                if ex is None:
                    ex = x - 1

                if (ox, y + 1) in points:
                    if x == ex + 1:
                        y += 1
                        x = ox
                    else:
                        y -= 1
                        break
                else:
                    if x <= ex:
                        y -= 1
                    break

        c_rect = pygame.Rect(
            ox * tilewidth,
            oy * tileheight,
            (ex - ox + 1) * tilewidth,
            (y - oy + 1) * tileheight,
        )

        rects.append(c_rect)

        rect = pygame.Rect(ox, oy, ex - ox + 1, y - oy + 1)
        points.difference_update({p for p in points if rect.collidepoint(p)})

        pick_rect(points, rects)

    pick_rect(point_set, rect_list)

    return rect_list



"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <https://www.gnu.org/licenses/>.

Property parsing and casting for pytmx.

This module provides the casting maps used to coerce XML attribute strings
into appropriate Python types, and the `parse_properties` function that
reads Tiled custom properties.
"""

from __future__ import annotations

import logging
from collections import defaultdict
from copy import deepcopy
from typing import Any, Optional
from xml.etree import ElementTree

from .utils import convert_to_bool

logger = logging.getLogger(__name__)

# used to change the unicode string returned from xml to
# proper python variable types.
types = defaultdict(lambda: str)

types.update(
    {
        "backgroundcolor": str,
        "bold": convert_to_bool,
        "color": str,
        "columns": int,
        "compression": str,
        "draworder": str,
        "duration": int,
        "encoding": str,
        "firstgid": int,
        "fontfamily": str,
        "format": str,
        "gid": int,
        "halign": str,
        "height": float,
        "hexsidelength": float,
        "id": int,
        "italic": convert_to_bool,
        "kerning": convert_to_bool,
        "margin": int,
        "name": str,
        "nextobjectid": int,
        "offsetx": int,
        "offsety": int,
        "opacity": float,
        "orientation": str,
        "pixelsize": float,
        "points": str,
        "probability": float,
        "renderorder": str,
        "rotation": float,
        "source": str,
        "spacing": int,
        "staggeraxis": str,
        "staggerindex": str,
        "strikeout": convert_to_bool,
        "terrain": str,
        "tile": int,
        "tilecount": int,
        "tiledversion": str,
        "tileheight": int,
        "tileid": int,
        "tilewidth": int,
        "trans": str,
        "type": str,
        "underline": convert_to_bool,
        "valign": str,
        "value": str,
        "version": str,
        "visible": convert_to_bool,
        "width": float,
        "wrap": convert_to_bool,
        "x": float,
        "y": float,
    }
)


def resolve_to_class(value: str, custom_types: dict) -> Any:
    """Convert Tiled custom type name to its defined Python object copy."""
    if value not in custom_types:
        raise ValueError(f"Custom type {value} not found.")
    return deepcopy(custom_types[value])


# casting for properties type
prop_type = {
    "bool": convert_to_bool,
    "color": str,
    "file": str,
    "float": float,
    "int": int,
    "object": int,
    "string": str,
    "class": resolve_to_class,
    "enum": str,
}


def parse_properties(
    node: ElementTree.Element, customs: Optional[dict[str, Any]] = None
) -> dict[Any, Any]:
    """Parse a Tiled XML node and return a property dict.

    This parses `<properties>` children and casts their values according to
    `prop_type`. If a property is of type `class`, it instantiates a new
    object (via `resolve_to_class`) and recursively assigns nested members.
    """
    d = dict()
    for child in node.findall("properties"):
        for subnode in child.findall("property"):
            cls = None
            try:
                if "type" in subnode.keys():
                    cls = prop_type[subnode.get("type")]
            except KeyError:
                logger.info(
                    f"Type {subnode.get('type')} not a built-in type. Defaulting to string-cast."
                )

            if "class" == subnode.get("type"):
                new = resolve_to_class(subnode.get("propertytype"), customs or {})
                properties = parse_properties(subnode, customs)
                for key in properties.keys():
                    setattr(new, key, properties[key])
                d[subnode.get("name")] = new
            else:
                d[subnode.get("name")] = subnode.get("value") or subnode.text
                if cls is not None:
                    d[subnode.get("name")] = cls(subnode.get("value"))
    return d



"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <https://www.gnu.org/licenses/>.

Tiled group layer model and parser.
"""

from typing import TYPE_CHECKING, Self
from xml.etree import ElementTree

from .element import TiledElement

if TYPE_CHECKING:
    from .map import TiledMap


class TiledGroupLayer(TiledElement):
    def __init__(self, parent: "TiledMap", node: ElementTree.Element) -> None:
        """

        Args:
            parent (TiledMap): The parent TiledMap.
            node (ElementTree.Element): The XML node to parse.
        """
        super().__init__()
        self.parent = parent
        self.name = None
        self.visible = 1
        self.parse_xml(node)

    def parse_xml(self, node: ElementTree.Element) -> Self:
        """
        Parse a TiledGroupLayer layer from ElementTree xml node.

        Returns:
            TiledGroupLayer: The parsed TiledGroupLayer layer.
        """
        self._set_properties(node)
        self.name = node.get("name", None)
        return self



"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <https://www.gnu.org/licenses/>.

Tiled tile layer model and parser.
"""

import logging
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Optional, Self
from xml.etree import ElementTree

from .chunk import extract_chunks, stitch_chunks
from .element import TiledElement
from .utils import reshape_data, unpack_gids

if TYPE_CHECKING:
    from .map import TiledMap

logger = logging.getLogger(__name__)


class TiledTileLayer(TiledElement):
    """Represents a TileLayer.

    To just get the tile images, use TiledTileLayer.tiles().
    """

    def __init__(self, parent: "TiledMap", node: ElementTree.Element) -> None:
        super().__init__()
        self.parent = parent
        self.data = []
        self.chunks = []

        # defaults from the specification
        self.name = None
        self.width = 0
        self.height = 0
        self.opacity = 1.0
        self.visible = True
        self.offsetx = 0
        self.offsety = 0

        self.parse_xml(node)

    def __iter__(self) -> Iterable[tuple[int, int, int]]:
        return self.iter_data()

    def iter_data(self) -> Iterable[tuple[int, int, int]]:
        """Yields X, Y, GID tuples for each tile in the layer.

        Returns:
            Iterable[Tuple[int, int, int]]: Iterator of X, Y, GID tuples for each tile in the layer.
        """
        for y, row in enumerate(self.data):
            for x, gid in enumerate(row):
                yield x, y, gid

    def tiles(self) -> Iterable[tuple[int, int, Any]]:
        """Yields X, Y, Image tuples for each tile in the layer.

        Yields:
            Iterable[Tuple[int, int, Any]]: Iterator of X, Y, Image tuples for each tile in the layer
        """
        images = self.parent.images
        for x, y, gid in [i for i in self.iter_data() if i[2]]:
            yield x, y, images[gid]

    def _set_properties(
        self, node: ElementTree.Element, customs: Optional[dict[str, Any]] = None
    ) -> None:
        super()._set_properties(node, customs)

        # TODO: make class/layer-specific type casting
        # layer height and width must be int, but TiledElement.set_properties()
        # make a float by default, so recast as int here
        self.height = int(self.height)
        self.width = int(self.width)

    def parse_xml(self, node: ElementTree.Element) -> Self:
        """
        Parse a TiledTileLayer layer from ElementTree xml node.

        Returns:
            TiledTileLayer: The parsed TiledTileLayer layer.
        """
        self._set_properties(node)
        data_node = node.find("data")
        chunk_nodes = data_node.findall("chunk")
        if chunk_nodes:
            self.chunks = extract_chunks(
                chunk_nodes,
                encoding=data_node.get("encoding"),
                compression=data_node.get("compression"),
            )
            self.data = stitch_chunks(self.chunks, self.width, self.height, self.parent)
            return self

        child = data_node.find("tile")
        if child is not None:
            raise ValueError(
                "XML tile elements are no longer supported. Must use base64 or csv map formats."
            )

        temp = [
            self.parent.register_gid_check_flags(gid)
            for gid in unpack_gids(
                text=data_node.text.strip(),
                encoding=data_node.get("encoding", None),
                compression=data_node.get("compression", None),
            )
        ]

        self.data = reshape_data(temp, self.width)
        return self



"""
Copyright (C) 2012-2024, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging

from .pytmx import *

logger = logging.getLogger(__name__)

try:
    from pytmx.util_pygame import load_pygame
except ImportError:
    logger.debug("cannot import pygame tools")

__version__ = (3, 32)



"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <https://www.gnu.org/licenses/>.

Utility functions for pytmx.

This module contains helper functions that are independent of the core
classes and can be reused across the package.
"""

from __future__ import annotations

import gzip
import struct
import zlib
from base64 import b64decode
from collections.abc import Sequence
from logging import getLogger
from math import cos, radians, sin
from typing import Optional, Union

logger = getLogger(__name__)

try:
    import zstd
except ImportError:
    logger.warning("zstd compression is not installed. Disabling zstd support.")
    zstd = None

from .constants import (
    GID_MASK,
    GID_TRANS_FLIPX,
    GID_TRANS_FLIPY,
    GID_TRANS_ROT,
    Point,
    TileFlags,
    empty_flags,
    flag_cache,
)


def default_image_loader(filename: str, flags, **kwargs):
    """Return a lazy image loader that carries filename, rect, and flags.

    This default loader allows loading a map without actually decoding images.
    It returns a callable that, when invoked, returns a tuple with the
    filename, the rectangle within the image (if any), and the flags
    requested by the caller.
    """

    def load(rect=None, flags=None):
        return filename, rect, flags

    return load


def decode_gid(raw_gid: int) -> tuple[int, TileFlags]:
    """Decode a GID from TMX data into a base GID and its transform flags."""
    if raw_gid < GID_TRANS_ROT:
        return raw_gid, empty_flags

    # Check if the GID is already in the cache
    if raw_gid in flag_cache:
        return raw_gid & ~GID_MASK, flag_cache[raw_gid]

    # Calculate and cache the flags
    flags = TileFlags(
        raw_gid & GID_TRANS_FLIPX == GID_TRANS_FLIPX,
        raw_gid & GID_TRANS_FLIPY == GID_TRANS_FLIPY,
        raw_gid & GID_TRANS_ROT == GID_TRANS_ROT,
    )
    flag_cache[raw_gid] = flags
    return raw_gid & ~GID_MASK, flags


def reshape_data(gids: list[int], width: int) -> list[list[int]]:
    """Change 1D list of GIDs to a 2D list with rows of given width."""
    return [gids[i : i + width] for i in range(0, len(gids), width)]


def unpack_gids(
    text: str,
    encoding: Optional[str] = None,
    compression: Optional[str] = None,
) -> list[int]:
    """Return all GIDs from encoded/compressed layer data."""
    if encoding == "base64":
        data = b64decode(text)
        if compression == "gzip":
            data = gzip.decompress(data)
        elif compression == "zlib":
            data = zlib.decompress(data)
        elif compression == "zstd":
            if zstd:
                data = zstd.decompress(data)
            else:
                raise ValueError("zstd compression is not installed.")
        elif compression:
            raise ValueError(f"layer compression {compression} is not supported.")
        fmt = "<%dL" % (len(data) // 4)
        return list(struct.unpack(fmt, data))
    elif encoding == "csv":
        if not text.strip():
            return []
        return [int(i) for i in text.split(",")]
    elif encoding:
        raise ValueError(f"layer encoding {encoding} is not supported.")
    else:
        # When no encoding is provided, Tiled may have inline <tile gid="..."/> entries;
        # the caller is responsible for that path and providing already-parsed integers.
        return []


def convert_to_bool(value: Optional[Union[str, int, float]] = None) -> bool:
    """Convert common text/number variants to a boolean value.

    Recognizes: 1, y, t, true, yes as True
                -, 0, n, f, false, no as False
    """
    value = str(value).strip()
    if value:
        value = value.lower()[0]
        if value in ("1", "y", "t", "true", "yes"):
            return True
        if value in ("-", "0", "n", "f", "false", "no"):
            return False
    else:
        return False
    raise ValueError(f'cannot parse "{value}" as bool')


def rotate(
    points: Sequence[Point],
    origin: Point,
    angle: Union[int, float],
) -> list[Point]:
    """Rotate a sequence of points around an origin by angle degrees."""
    sin_t = sin(radians(angle))
    cos_t = cos(radians(angle))
    new_points = list()
    for point in points:
        p = (
            origin.x + (cos_t * (point.x - origin.x) - sin_t * (point.y - origin.y)),
            origin.y + (sin_t * (point.x - origin.x) + cos_t * (point.y - origin.y)),
        )
        new_points.append(p)
    return new_points


def decode_chunk_data(
    text: str, encoding: Optional[str], compression: Optional[str]
) -> tuple[list[int], bytes]:
    """
    Decode and decompress chunk data from a Tiled map.

    Args:
        text: The raw text content of the chunk.
        encoding: The encoding format (e.g., "base64", "csv").
        compression: The compression method (e.g., "zlib", "gzip", "zstd").

    Returns:
        A tuple of (gids, raw_data), where:
            - gids is a list of tile GIDs
            - raw_data is the binary representation used to unpack GIDs
    """
    if encoding == "base64":
        raw_data = b64decode(text.strip())

        if compression == "zlib":
            raw_data = zlib.decompress(raw_data)
        elif compression == "gzip":
            raw_data = gzip.decompress(raw_data)
        elif compression == "zstd":
            if zstd:
                raw_data = zstd.decompress(raw_data)
            else:
                raise ValueError("zstd compression is not installed.")
        elif compression:
            raise ValueError(f"Unsupported compression: {compression}")

        fmt = "<%dL" % (len(raw_data) // 4)
        gids = list(struct.unpack(fmt, raw_data))

    elif encoding == "csv":
        gids = [int(i) for i in text.strip().split(",")]
        raw_data = b""  # CSV has no binary representation

    elif encoding:
        raise ValueError(f"Unsupported encoding: {encoding}")

    else:
        gids = []
        raw_data = b""

    return gids, raw_data



"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <https://www.gnu.org/licenses/>.

Utility functions for pytmx.

This module contains helper functions that are independent of the core
classes and can be reused across the package.
"""

from __future__ import annotations

from dataclasses import dataclass
from logging import getLogger
from typing import TYPE_CHECKING, Optional
from xml.etree import ElementTree

from .utils import decode_chunk_data

if TYPE_CHECKING:
    from .map import TiledMap

logger = getLogger(__name__)


@dataclass
class Chunk:
    position: tuple[int, int]  # (x, y) tile coordinates
    size: tuple[int, int]  # (width, height) in tiles
    grid: list[list[int]]  # 2D array of tile GIDs
    raw: bytes  # Raw decompressed binary data


def extract_chunks(
    chunk_nodes: list[ElementTree.Element],
    encoding: Optional[str],
    compression: Optional[str],
) -> list[Chunk]:
    """
    Extracts chunk data from a list of <chunk> XML nodes, using the specified encoding and compression.

    Args:
        chunk_nodes: List of <chunk> elements from a TMX file.
        encoding: The encoding format used for the chunk data (e.g., "base64", "csv").
        compression: The compression method applied to the chunk data (e.g., "zlib", "gzip", "zstd").

    Returns:
        list[Chunk]: List of Chunk objects containing decoded tile GIDs and raw binary data.
    """
    chunks: list[Chunk] = []

    for i, chunk in enumerate(chunk_nodes):
        x = int(chunk.get("x") or 0)
        y = int(chunk.get("y") or 0)
        width = int(chunk.get("width") or 0)
        height = int(chunk.get("height") or 0)

        logger.debug(f"[Chunk {i}] Position: ({x}, {y}), Size: {width}x{height}")

        if chunk.text is None:
            logger.error(f"[Chunk {i}] Missing text content in chunk")
            continue

        try:
            gids, raw_data = decode_chunk_data(
                text=chunk.text.strip(),
                encoding=encoding,
                compression=compression,
            )
        except Exception as e:
            logger.error(f"[Chunk {i}] Failed to decode GIDs: {e}")
            continue

        if len(gids) != width * height:
            logger.warning(
                f"[Chunk {i}] GID count mismatch: expected {width * height}, got {len(gids)}"
            )
        if len(gids) != width * height:
            logger.warning(
                f"[Chunk {i}] GID count mismatch: expected {width * height}, got {len(gids)}"
            )
        grid = [gids[row * width : (row + 1) * width] for row in range(height)]

        logger.debug(f"[Chunk {i}] Grid extracted with {len(grid)} rows")

        chunks.append(
            Chunk(position=(x, y), size=(width, height), grid=grid, raw=raw_data)
        )

    logger.info(f"Total chunks extracted: {len(chunks)}")
    return chunks


def stitch_chunks(
    chunks: list[Chunk], width: int, height: int, parent: TiledMap
) -> list[list[int]]:
    """
    Stitch together multiple chunks into a full tile grid, normalizing GIDs.

    Args:
        chunks: List of Chunk objects.
        width: Width of the full map in tiles.
        height: Height of the full map in tiles.
        parent: Reference to the TiledMap for GID normalization.

    Returns:
        A 2D list representing the full tile grid.
    """
    full_grid = [[0 for _ in range(width)] for _ in range(height)]

    for chunk_index, chunk in enumerate(chunks):
        cx, cy = chunk.position
        if cx < 0 or cy < 0:
            logger.warning(f"Skipping chunk at negative position ({cx}, {cy})")
            continue

        cw, ch = chunk.size
        out_of_bounds_logged = False

        for y in range(ch):
            for x in range(cw):
                raw_gid = chunk.grid[y][x]
                normalized_gid = parent.register_gid_check_flags(raw_gid)

                gx, gy = cx + x, cy + y
                if 0 <= gx < width and 0 <= gy < height:
                    full_grid[gy][gx] = normalized_gid
                elif not out_of_bounds_logged:
                    logger.warning(
                        f"[Chunk {chunk_index}] Contains out-of-bounds tiles (e.g., ({gx}, {gy}))"
                    )
                    out_of_bounds_logged = True

    logger.info("Chunks stitched successfully into full grid")
    return full_grid



import timeit

from pytmx.constants import TileFlags
from pytmx.utils import decode_gid

# Constants
GID_TRANS_FLIPX = 1 << 31
GID_TRANS_FLIPY = 1 << 30
GID_TRANS_ROT = 1 << 29
GID_MASK = GID_TRANS_FLIPX | GID_TRANS_FLIPY | GID_TRANS_ROT


def decode_no_cache(raw_gid: int) -> tuple[int, TileFlags]:
    if raw_gid < GID_TRANS_ROT:
        return raw_gid, TileFlags(False, False, False)
    return (
        raw_gid & ~GID_MASK,
        TileFlags(
            raw_gid & GID_TRANS_FLIPX == GID_TRANS_FLIPX,
            raw_gid & GID_TRANS_FLIPY == GID_TRANS_FLIPY,
            raw_gid & GID_TRANS_ROT == GID_TRANS_ROT,
        ),
    )


def run_benchmark(size: int) -> tuple[float, float]:
    raw_gids = [GID_TRANS_FLIPX | GID_TRANS_FLIPY | (i % 1000 + 1) for i in range(size)]

    def run_no_cache():
        for gid in raw_gids:
            decode_no_cache(gid)

    def run_with_cache():
        for gid in raw_gids:
            decode_gid(gid)

    time_no_cache = timeit.timeit(run_no_cache, number=1)
    time_cache = timeit.timeit(run_with_cache, number=1)
    return time_no_cache, time_cache


def main():
    print("Running decode_gid benchmarks...\n")
    gid_sizes = [1_000, 5_000, 10_000, 50_000, 100_000, 250_000, 500_000, 1_000_000]

    print(f"{'GID Count':>10} | {'No Cache':>10} | {'With Cache':>11} | {'Speedup':>8}")
    print("-" * 50)

    for size in gid_sizes:
        no_cache, cache = run_benchmark(size)
        speedup = no_cache / cache if cache > 0 else float("inf")
        print(f"{size:>10} | {no_cache:.4f}s | {cache:.4f}s | {speedup:>7.2f}x")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <http://www.gnu.org/licenses/>.
"""
import logging
from functools import partial
from typing import Any, Callable, Optional

from .constants import ColorLike, TileFlags
from .map import TiledMap

logger = logging.getLogger(__name__)

try:
    import sdl2.ext
except ImportError:
    logger.error("cannot import pysdl2 (is it installed?)")
    raise

__all__ = [
    "load_pysdl2",
    "pysdl2_image_loader",
]


def pysdl2_image_loader(
    renderer: sdl2.SDL_Renderer,
    filename: str,
    colorkey: Optional[ColorLike] = None,
    **kwargs: Any,
) -> Callable[[Optional[tuple[int, int, int, int]], Optional[TileFlags]], Any]:
    def convert(surface):
        texture_ = sdl2.SDL_CreateTextureFromSurface(renderer.renderer, surface)
        sdl2.SDL_SetTextureBlendMode(texture_, sdl2.SDL_BLENDMODE_BLEND)
        sdl2.SDL_FreeSurface(surface)
        return texture_

    def load_image(
        rect: Optional[tuple[int, int, int, int]] = None,
        flags: Optional[TileFlags] = None,
    ) -> Any:
        if rect:
            try:
                flip = 0
                if flags.flipped_horizontally:
                    flip |= sdl2.SDL_FLIP_HORIZONTAL
                if flags.flipped_vertically:
                    flip |= sdl2.SDL_FLIP_VERTICAL
                if flags.flipped_diagonally:
                    flip |= 4

                rect = sdl2.rect.SDL_Rect(*rect)
                return texture, rect, flip

            except ValueError:
                logger.error("Tile bounds outside bounds of tileset image")
                raise
        else:
            return texture, None, 0

    image = sdl2.ext.load_image(filename)

    if colorkey:
        colorkey = sdl2.ext.string_to_color("#" + colorkey)
        key = sdl2.SDL_MapRGB(image.format, *colorkey[:3])
        sdl2.SDL_SetColorKey(image, sdl2.SDL_TRUE, key)

    texture = convert(image)

    return load_image


def load_pysdl2(
    renderer: sdl2.SDL_Renderer, filename: str, *args: Any, **kwargs: Any
) -> TiledMap:
    kwargs["image_loader"] = partial(pysdl2_image_loader, renderer)
    return TiledMap(filename, *args, **kwargs)



"""
This is tested on pysdl2 1.2 and python 2.7.
Leif Theden "bitcraft", 2012-2024

Rendering demo for the TMXLoader.

This should be considered --alpha-- quality.  I'm including it as a
proof-of-concept for now and will improve on it in the future.

For Windows users, you will need to download SDL2 runtime and
place it where pysdl2 can find it.  Please check the pysdl2 docs
for information on that process.

Notice: slow!  no transparency!  no tile rotation!
"""

import logging

logger = logging.getLogger(__name__)
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
logger.addHandler(ch)
logger.setLevel(logging.INFO)

# QUICK SDL2 HACK FOR WINDOWS
# 1. download and move SDL2.dll to apps folder
# 2. uncomment the two lines of code below
# 3. profit!
import os

os.environ["PYSDL2_DLL_PATH"] = os.path.dirname(__file__)

import sdl2.ext
from sdl2 import *

from pytmx import *
from pytmx.util_pysdl2 import load_pysdl2


class TiledRenderer:
    """
    Super simple way to render a tiled map with pyglet

    no shape drawing yet
    """

    def __init__(self, filename, renderer) -> None:
        tm = load_pysdl2(renderer, filename)
        self.size = tm.width * tm.tilewidth, tm.height * tm.tileheight
        self.tmx_data = tm
        self.renderer = renderer

    def render_tile_layer(self, layer) -> None:
        """Render the tile layer

        DOES NOT CHECK FOR DRAWING TILES OFF THE SCREEN
        """
        # deref these heavily used references for speed
        tw = self.tmx_data.tilewidth
        th = self.tmx_data.tileheight
        renderer = self.renderer.renderer
        dest = sdl2.rect.SDL_Rect(0, 0, tw, th)
        rce = sdl2.SDL_RenderCopyEx

        # iterate over the tiles in the layer
        for x, y, tile in layer.tiles():
            texture, src, flip = tile
            dest.x = x * tw
            dest.y = y * th
            angle = 90 if (flip & 4) else 0
            rce(renderer, texture, src, dest, angle, None, flip)

    def render_map(self) -> None:
        """Render the entire map

        Only tile layer drawing is implemented
        """
        for layer in self.tmx_data.visible_layers:

            # draw map tile layers
            if isinstance(layer, TiledTileLayer):
                self.render_tile_layer(layer)


class SimpleTest:
    def __init__(self, filename, window) -> None:
        self.running = False
        self.dirty = False
        self.exit_status = 0
        self.sdl_renderer = window.renderer
        self.map_renderer = TiledRenderer(filename, self.sdl_renderer)

        logger.info("Objects in map:")
        for obj in self.map_renderer.tmx_data.objects:
            logger.info(obj)
            for k, v in obj.properties.items():
                logger.info("%s\t%s", k, v)

        logger.info("GID (tile) properties:")
        for k, v in self.map_renderer.tmx_data.tile_properties.items():
            logger.info("%s\t%s", k, v)

    def draw(self) -> None:
        self.sdl_renderer.clear()
        self.map_renderer.render_map()
        self.sdl_renderer.present()

    def run(self, window):
        """Starts an event loop without actually processing any event."""
        import ctypes

        event = events.SDL_Event()
        self.running = True
        self.exit_status = 1
        while self.running:
            ret = events.SDL_PollEvent(ctypes.byref(event), 1)
            if ret == 1:
                if event.type == SDL_QUIT:
                    self.exit_status = 0
                    self.running = False
                elif event.type == SDL_KEYDOWN:
                    self.running = False
            self.draw()
            window.refresh()
            timer.SDL_Delay(10)

        return self.exit_status


def all_filenames():
    import glob
    import os.path

    return glob.glob(os.path.join("data", "*.tmx"))


if __name__ == "__main__":
    window = sdl2.ext.Window("pytmx + psdl2 = awesome???", size=(600, 600))
    window.renderer = sdl2.ext.Renderer(window)
    window.renderer.blendmode = SDL_BLENDMODE_BLEND
    window.renderer.color = 0, 0, 0, 0
    window.show()

    try:
        for filename in all_filenames():
            logger.info("Testing %s", filename)
            if not SimpleTest(filename, window).run(window):
                break
    except:
        raise



"""
Copyright (C) 2012-2025, Leif Theden <leif.theden@gmail.com>

This file is part of pytmx.

pytmx is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

pytmx is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with pytmx.  If not, see <https://www.gnu.org/licenses/>.

"""

from .class_type import TiledClassType
from .constants import TileFlags
from .element import TiledElement
from .group_layer import TiledGroupLayer
from .image_layer import TiledImageLayer
from .map import TiledMap
from .object import TiledObject
from .object_group import TiledObjectGroup
from .properties import parse_properties, resolve_to_class
from .property import TiledProperty
from .tile_layer import TiledTileLayer
from .tileset import TiledTileset
from .utils import convert_to_bool, decode_gid, unpack_gids

__all__ = (
    "TileFlags",
    "TiledElement",
    "TiledImageLayer",
    "TiledMap",
    "TiledGroupLayer",
    "TiledObject",
    "TiledObjectGroup",
    "TiledTileLayer",
    "TiledProperty",
    "TiledClassType",
    "TiledTileset",
    "convert_to_bool",
    "resolve_to_class",
    "parse_properties",
    "decode_gid",
    "unpack_gids",
)



import base64
import struct
import unittest
import xml.etree.ElementTree as ET
import zlib
from unittest.mock import MagicMock

from pytmx.chunk import Chunk, extract_chunks, stitch_chunks


class TestExtractChunks(unittest.TestCase):
    def setUp(self):
        self.gids = [1, 2, 3, 4]
        self.width = 2
        self.height = 2
        self.encoding = "base64"
        self.compression = "zlib"

        packed = struct.pack("<4I", *self.gids)
        compressed = zlib.compress(packed)
        encoded = base64.b64encode(compressed).decode("ascii")

        self.chunk_xml = ET.Element(
            "chunk",
            {"x": "0", "y": "0", "width": str(self.width), "height": str(self.height)},
        )
        self.chunk_xml.text = encoded

    def test_valid_chunk_extraction(self):
        chunks = extract_chunks(
            [self.chunk_xml], encoding=self.encoding, compression=self.compression
        )
        self.assertEqual(len(chunks), 1)

        chunk = chunks[0]
        self.assertEqual(chunk.position, (0, 0))
        self.assertEqual(chunk.size, (2, 2))
        self.assertEqual(chunk.grid, [[1, 2], [3, 4]])
        self.assertEqual(
            chunk.raw, zlib.decompress(base64.b64decode(self.chunk_xml.text))
        )

    def test_byte_mismatch_warning(self):
        gids = [1, 2]  # Only 2 GIDs, but chunk expects 2x2 = 4
        packed = struct.pack("<2I", *gids)
        compressed = zlib.compress(packed)
        corrupted = base64.b64encode(compressed).decode("ascii")
        corrupted_chunk = ET.Element(
            "chunk",
            {"x": "0", "y": "0", "width": str(self.width), "height": str(self.height)},
        )
        corrupted_chunk.text = corrupted

        with self.assertLogs("pytmx.chunk", level="WARNING") as cm:
            extract_chunks([corrupted_chunk], encoding="base64", compression="zlib")

        self.assertTrue(any("GID count mismatch" in msg for msg in cm.output))

    def test_multiple_chunks(self):
        chunk2 = ET.Element(
            "chunk",
            {"x": "2", "y": "0", "width": str(self.width), "height": str(self.height)},
        )
        chunk2.text = self.chunk_xml.text

        chunks = extract_chunks(
            [self.chunk_xml, chunk2],
            encoding=self.encoding,
            compression=self.compression,
        )
        self.assertEqual(len(chunks), 2)
        self.assertEqual(chunks[1].position, (2, 0))
        self.assertEqual(chunks[1].grid, [[1, 2], [3, 4]])

    def test_invalid_attributes(self):
        bad_chunk = ET.Element(
            "chunk", {"x": "not-an-int", "y": "0", "width": "2", "height": "2"}
        )
        bad_chunk.text = self.chunk_xml.text

        with self.assertRaises(ValueError):
            extract_chunks(
                [bad_chunk], encoding=self.encoding, compression=self.compression
            )

    def test_malformed_base64(self):
        malformed_chunk = ET.Element(
            "chunk", {"x": "0", "y": "0", "width": "2", "height": "2"}
        )
        malformed_chunk.text = "!!!notbase64!!!"

        with self.assertLogs("pytmx.chunk", level="ERROR") as cm:
            extract_chunks([malformed_chunk], encoding="base64", compression="zlib")

        self.assertTrue(any("Failed to decode GIDs" in msg for msg in cm.output))


class TestStitchChunks(unittest.TestCase):
    def setUp(self):
        self.mock_map = MagicMock()
        self.mock_map.register_gid_check_flags.side_effect = (
            lambda gid: gid & 0x1FFFFFFF
        )  # Strip flip flags

        self.chunk1 = Chunk(
            position=(0, 0),
            size=(2, 2),
            grid=[[1, 2], [3, 4 | 0x80000000]],  # flipped tile
            raw=b"",  # not used in stitching
        )

        self.chunk2 = Chunk(
            position=(2, 0),
            size=(2, 2),
            grid=[[5, 6], [7 | 0x40000000, 8]],  # flipped tile
            raw=b"",
        )

        self.chunks = [self.chunk1, self.chunk2]
        self.width = 4
        self.height = 2

    def test_stitch_chunks_correct_grid(self):
        expected_grid = [[1, 2, 5, 6], [3, 4, 7, 8]]

        result = stitch_chunks(self.chunks, self.width, self.height, self.mock_map)
        self.assertEqual(result, expected_grid)

    def test_gid_normalization_called(self):
        stitch_chunks(self.chunks, self.width, self.height, self.mock_map)

        # Flatten all GIDs from both chunks
        raw_gids = [gid for chunk in self.chunks for row in chunk.grid for gid in row]
        normalized_calls = [((gid,),) for gid in raw_gids]

        # Check that register_gid_check_flags was called with each raw GID
        self.mock_map.register_gid_check_flags.assert_has_calls(
            normalized_calls, any_order=True
        )

    def test_out_of_bounds_tile_skipped(self):
        # Add a chunk that goes out of bounds
        out_of_bounds_chunk = Chunk(
            position=(3, 1), size=(2, 2), grid=[[9, 10], [11, 12]], raw=b""
        )
        chunks = self.chunks + [out_of_bounds_chunk]

        result = stitch_chunks(chunks, self.width, self.height, self.mock_map)

        # Ensure out-of-bounds tiles are not written
        self.assertEqual(result[1][3], 9)  # overwritten by out-of-bounds chunk
        # No IndexError should occur

    def test_empty_chunks(self):
        result = stitch_chunks([], self.width, self.height, self.mock_map)
        expected = [[0, 0, 0, 0], [0, 0, 0, 0]]
        self.assertEqual(result, expected)

    def test_overlapping_chunks_last_write_wins(self):
        overlapping_chunk = Chunk(
            position=(1, 0), size=(2, 2), grid=[[100, 101], [102, 103]], raw=b""
        )
        chunks = [self.chunk1, overlapping_chunk]

        result = stitch_chunks(chunks, self.width, self.height, self.mock_map)

        # Expect overlapping_chunk to overwrite chunk1 at (1,0) and (2,0)
        self.assertEqual(result[0][1], 100)
        self.assertEqual(result[0][2], 101)

    def test_negative_position_chunk(self):
        negative_chunk = Chunk(
            position=(-1, -1), size=(2, 2), grid=[[200, 201], [202, 203]], raw=b""
        )
        chunks = [negative_chunk]

        result = stitch_chunks(chunks, self.width, self.height, self.mock_map)

        # Nothing should be written to the grid
        expected = [[0, 0, 0, 0], [0, 0, 0, 0]]
        self.assertEqual(result, expected)

    def test_partially_out_of_bounds_chunk(self):
        partial_chunk = Chunk(
            position=(3, 1), size=(2, 2), grid=[[300, 301], [302, 303]], raw=b""
        )
        chunks = [partial_chunk]

        result = stitch_chunks(chunks, self.width, self.height, self.mock_map)

        # Only (3,1) should be written
        self.assertEqual(result[1][3], 300)

    def test_gid_normalization_logic(self):
        raw_gid = 0x80000001
        normalized = self.mock_map.register_gid_check_flags(raw_gid)
        self.assertEqual(normalized, 1)

    def test_irregular_chunk_grid(self):
        irregular_chunk = Chunk(
            position=(0, 0), size=(2, 2), grid=[[1], [2, 3]], raw=b""  # too short
        )
        chunks = [irregular_chunk]

        with self.assertRaises(IndexError):
            stitch_chunks(chunks, self.width, self.height, self.mock_map)



import unittest
from unittest.mock import MagicMock, patch
from xml.etree.ElementTree import Element, SubElement

from pytmx.constants import AnimationFrame
from pytmx.tileset import TiledTileset


class TestTiledTileset(unittest.TestCase):

    def setUp(self):
        self.mock_parent = MagicMock()
        self.mock_parent.filename = "/maps/test_map.tmx"
        self.mock_parent.register_gid = lambda gid: gid + 1000
        self.mock_parent.map_gid2 = lambda gid: [(gid, 0)]
        self.mock_parent.set_tile_properties = MagicMock()

    def create_basic_tileset_node(self):
        node = Element(
            "tileset",
            {
                "name": "TestTileset",
                "tilewidth": "32",
                "tileheight": "32",
                "tilecount": "4",
                "columns": "2",
            },
        )
        image = SubElement(
            node, "image", {"source": "tiles.png", "width": "64", "height": "64"}
        )
        return node

    def test_basic_tileset_parsing(self):
        node = self.create_basic_tileset_node()
        tileset = TiledTileset(self.mock_parent, node)

        self.assertEqual(tileset.name, "TestTileset")
        self.assertEqual(tileset.tilewidth, 32)
        self.assertEqual(tileset.tileheight, 32)
        self.assertEqual(tileset.tilecount, 4)
        self.assertEqual(tileset.columns, 2)
        self.assertEqual(tileset.source, "tiles.png")
        self.assertEqual(tileset.width, 64)
        self.assertEqual(tileset.height, 64)

    def test_external_tileset_loading(self):
        tsx_path = "/maps/test_tileset.tsx"
        tsx_node = self.create_basic_tileset_node()

        with (
            patch("os.path.exists", return_value=True),
            patch("xml.etree.ElementTree.parse") as mock_parse,
        ):
            mock_parse.return_value.getroot.return_value = tsx_node

            external_node = Element(
                "tileset", {"firstgid": "1", "source": "test_tileset.tsx"}
            )
            tileset = TiledTileset(self.mock_parent, external_node)

            self.assertEqual(tileset.firstgid, 1)
            self.assertEqual(tileset.name, "TestTileset")

    def test_missing_external_file_raises(self):
        external_node = Element("tileset", {"firstgid": "1", "source": "missing.tsx"})

        with patch("os.path.exists", return_value=False):
            with self.assertRaises(Exception) as context:
                TiledTileset(self.mock_parent, external_node)
            self.assertIn("Cannot find tileset file", str(context.exception))

    def test_tile_with_animation(self):
        node = self.create_basic_tileset_node()
        tile = SubElement(node, "tile", {"id": "0"})
        anim = SubElement(tile, "animation")
        SubElement(anim, "frame", {"tileid": "1", "duration": "100"})
        SubElement(anim, "frame", {"tileid": "2", "duration": "200"})

        tileset = TiledTileset(self.mock_parent, node)
        props = self.mock_parent.set_tile_properties.call_args[0][1]

        self.assertIn("frames", props)
        self.assertEqual(len(props["frames"]), 2)
        self.assertIsInstance(props["frames"][0], AnimationFrame)

    def test_tile_with_image_and_transparency(self):
        node = self.create_basic_tileset_node()
        tile = SubElement(node, "tile", {"id": "0"})
        image = SubElement(
            tile,
            "image",
            {"source": "tile0.png", "trans": "ff00ff", "width": "32", "height": "32"},
        )

        tileset = TiledTileset(self.mock_parent, node)
        props = self.mock_parent.set_tile_properties.call_args[0][1]

        self.assertEqual(props["source"], "tile0.png")
        self.assertEqual(props["trans"], "ff00ff")
        self.assertEqual(props["width"], 32)
        self.assertEqual(props["height"], 32)

    def test_tileoffset_parsing(self):
        node = self.create_basic_tileset_node()
        offset = SubElement(node, "tileoffset", {"x": "5", "y": "10"})

        tileset = TiledTileset(self.mock_parent, node)
        self.assertEqual(tileset.offset, (5, 10))

    def test_tile_without_image_defaults(self):
        node = self.create_basic_tileset_node()
        tile = SubElement(node, "tile", {"id": "0"})

        tileset = TiledTileset(self.mock_parent, node)
        props = self.mock_parent.set_tile_properties.call_args[0][1]

        self.assertEqual(props["width"], 32)
        self.assertEqual(props["height"], 32)

    def test_empty_tileset_node(self):
        node = Element("tileset")
        tileset = TiledTileset(self.mock_parent, node)
        self.assertEqual(tileset.name, None)
        self.assertEqual(tileset.tilewidth, 0)

    def test_invalid_tilewidth_type(self):
        node = Element("tileset", {"tilewidth": "not_an_int"})
        with self.assertRaises(ValueError):
            TiledTileset(self.mock_parent, node)

    def test_multiple_tileoffset_nodes(self):
        node = self.create_basic_tileset_node()
        SubElement(node, "tileoffset", {"x": "1", "y": "2"})
        SubElement(node, "tileoffset", {"x": "99", "y": "99"})  # Should be ignored

        tileset = TiledTileset(self.mock_parent, node)
        self.assertEqual(tileset.offset, (1, 2))

    def test_tile_missing_id(self):
        node = self.create_basic_tileset_node()
        tile = SubElement(node, "tile")  # No 'id'

        with self.assertRaises(TypeError):
            TiledTileset(self.mock_parent, node)

    def test_tileset_with_unknown_tags(self):
        node = self.create_basic_tileset_node()
        SubElement(node, "unknownTag", {"foo": "bar"})

        tileset = TiledTileset(self.mock_parent, node)
        self.assertEqual(tileset.name, "TestTileset")  # Still parses correctly

    def test_tileset_without_image(self):
        node = Element(
            "tileset", {"name": "NoImageTileset", "tilewidth": "32", "tileheight": "32"}
        )

        tileset = TiledTileset(self.mock_parent, node)
        self.assertIsNone(tileset.source)
        self.assertEqual(tileset.width, 0)

    def test_tile_with_objectgroup(self):
        node = self.create_basic_tileset_node()
        tile = SubElement(node, "tile", {"id": "0"})
        objgrp = SubElement(tile, "objectgroup")
        SubElement(objgrp, "object", {"id": "1", "x": "10", "y": "20"})

        tileset = TiledTileset(self.mock_parent, node)
        props = self.mock_parent.set_tile_properties.call_args[0][1]
        self.assertIn("colliders", props)



import unittest
from unittest.mock import patch
from xml.etree.ElementTree import Element

from pytmx.element import TiledElement
from pytmx.properties import parse_properties


class DummyElement(TiledElement):
    def __init__(self, allow_duplicate_names=False):
        super().__init__(allow_duplicate_names=allow_duplicate_names)

    def parse_xml(self, node):
        self._set_properties(node)
        return self


class CustomClass(TiledElement):
    def __init__(self):
        super().__init__()
        self.properties["foo"] = "bar"

    def parse_xml(self, node):
        self._set_properties(node)
        return self


class TestTiledElement(unittest.TestCase):

    def setUp(self):
        self.element = DummyElement()

    def test_initial_state(self):
        self.assertFalse(self.element._allow_duplicate_names)
        self.assertEqual(self.element.properties, {})

    def test_cast_and_set_attributes(self):
        items = [("width", "32"), ("height", "64"), ("visible", "1")]
        self.element._cast_and_set_attributes_from_node_items(items)
        self.assertEqual(self.element.width, 32.0)
        self.assertEqual(self.element.height, 64.0)
        self.assertTrue(self.element.visible)

    def test_invalid_property_name_detection(self):
        self.element.name = "TestElement"
        items = [("name", "conflict")]
        self.assertTrue(self.element._contains_invalid_property_name(items))

    def test_valid_property_name_with_duplicates_allowed(self):
        self.element.allow_duplicate_names = True
        items = [("name", "conflict")]
        self.assertFalse(self.element._contains_invalid_property_name(items))

    def test_set_properties_with_valid_data(self):
        xml = Element("element", attrib={"width": "128", "height": "256"})
        props = Element("properties")
        prop = Element("property", attrib={"name": "custom", "value": "hello"})
        props.append(prop)
        xml.append(props)

        self.element._set_properties(xml)
        self.assertEqual(self.element.width, 128.0)
        self.assertEqual(self.element.height, 256.0)
        self.assertEqual(self.element.properties["custom"], "hello")

    def test_set_properties_with_conflict_raises(self):
        xml = Element("element", attrib={"name": "conflict"})
        props = Element("properties")
        prop = Element("property", attrib={"name": "name", "value": "oops"})
        props.append(prop)
        xml.append(props)

        with self.assertRaises(ValueError):
            self.element._set_properties(xml)

    def test_getattr_existing_property(self):
        self.element.properties["foo"] = "bar"
        self.assertEqual(self.element.foo, "bar")

    def test_getattr_missing_property_with_name(self):
        self.element.properties["name"] = "TestElement"
        with self.assertRaises(AttributeError) as cm:
            _ = self.element.missing
        self.assertIn("TestElement", str(cm.exception))

    def test_getattr_missing_property_without_name(self):
        with self.assertRaises(AttributeError) as cm:
            _ = self.element.missing
        self.assertIn("Element has no property", str(cm.exception))

    def test_repr_with_id(self):
        self.element.id = 42
        self.element.name = "MyElement"
        self.assertEqual(repr(self.element), '<DummyElement[42]: "MyElement">')

    def test_repr_without_id(self):
        self.element.name = "MyElement"
        self.assertEqual(repr(self.element), '<DummyElement: "MyElement">')

    def test_from_xml_string(self):
        xml = """
        <element width="100" height="200">
            <properties>
                <property name="custom" value="value" />
            </properties>
        </element>
        """
        obj = DummyElement.from_xml_string(xml)
        self.assertEqual(obj.width, 100.0)
        self.assertEqual(obj.height, 200.0)
        self.assertEqual(obj.properties["custom"], "value")

    def test_property_type_casting(self):
        xml = Element("element")
        props = Element("properties")
        prop = Element(
            "property", attrib={"name": "visible", "type": "bool", "value": "true"}
        )
        props.append(prop)
        xml.append(props)

        props_dict = parse_properties(xml)
        self.assertTrue(props_dict["visible"])

    @patch("pytmx.properties.deepcopy", lambda x: x)
    def test_nested_class_property(self):
        customs = {"MyClass": CustomClass()}
        xml = Element("element")
        props = Element("properties")
        class_prop = Element(
            "property",
            attrib={"name": "nested", "type": "class", "propertytype": "MyClass"},
        )
        subprop = Element("property", attrib={"name": "foo", "value": "bar"})
        class_prop.append(subprop)
        props.append(class_prop)
        xml.append(props)

        props_dict = parse_properties(xml, customs)
        self.assertEqual(props_dict["nested"].foo, "bar")

    def test_property_fallback_to_text(self):
        xml = Element("element")
        props = Element("properties")
        prop = Element("property", attrib={"name": "fallback"})
        prop.text = "fallback_value"
        props.append(prop)
        xml.append(props)

        props_dict = parse_properties(xml)
        self.assertEqual(props_dict["fallback"], "fallback_value")

    def test_property_type_not_found_logs_info(self):
        xml = Element("element")
        props = Element("properties")
        prop = Element(
            "property", attrib={"name": "unknown", "type": "mystery", "value": "42"}
        )
        props.append(prop)
        xml.append(props)

        props_dict = parse_properties(xml)
        self.assertEqual(props_dict["unknown"], "42")

    def test_parse_xml_sets_properties(self):
        xml = Element("element", attrib={"width": "100", "height": "200"})
        props = Element("properties")
        prop = Element("property", attrib={"name": "custom", "value": "hello"})
        props.append(prop)
        xml.append(props)

        result = self.element.parse_xml(xml)
        self.assertEqual(result.width, 100.0)
        self.assertEqual(result.height, 200.0)
        self.assertEqual(result.properties["custom"], "hello")

    def test_repr_output(self):
        self.element.id = 1
        self.element.name = "TestDummy"
        self.assertEqual(repr(self.element), '<DummyElement[1]: "TestDummy">')

    def test_property_access(self):
        self.element.properties["foo"] = "bar"
        self.assertEqual(self.element.foo, "bar")

    def test_missing_property_raises(self):
        with self.assertRaises(AttributeError):
            _ = self.element.nonexistent

    def test_contains_invalid_property_name(self):
        self.element.name = "dummy"
        items = [("name", "conflict")]
        self.assertTrue(self.element._contains_invalid_property_name(items))
